#ifdef PETSC_RCS_HEADER
static char vcid[] = "$Id: navierstokes.c,v 1.8 1999/09/13 10:48:25 knepley Exp $";
#endif

static char help[] = "This is the Navier-Stokes problem with P2/P1 elements.\n\n";

int NAVIERSTOKES_COOKIE;
int NAVIERSTOKES_ComputeField;

#include "navierstokes.h"

/*
  Here we are solving the Navier-Stokes equation:

    / u \cdot \nabla u - \nu \Delta u - \nabla p \ = / f(x)\
    \             \nabla \cdot u                 /   \  0  /

  We then make the associations:

      F(u,p) = / u \cdot \nabla u - \nu \Delta u - \nabla p - f(x) \
               \              \nabla \cdot u                       /
      J(u,p) = / LC(u) - \nu \Delta - \nabla \
               \        \nabla \cdot         /

  Thus we now let

    VelocitySolutionFunction() --> u_S (which we prescribe)
    PressureSolutionFunction() --> p_S (which we prescribe)
    VelocityRhsFunction()      --> -f
    ConvectionOperator()       --> u \cdot \nabla u
    LinearConvectionOperator() --> Linear form of convection
*/
#undef  __FUNCT__
#define __FUNCT__ "main"
int main(int argc, char **args)
{
  NavierStokesContext ctx; /* Holds problem specific information */
  int                 ierr;

  PetscFunctionBegin;
  ierr = PetscInitialize(&argc, &args, 0, help);                                    CHKERRABORT(PETSC_COMM_WORLD, ierr);

  ierr = NavierStokesInitializePackage(PETSC_NULL);                                 CHKERRABORT(PETSC_COMM_WORLD, ierr);
  ierr = NavierStokesContextCreate(PETSC_COMM_WORLD, &ctx);                         CHKERRABORT(PETSC_COMM_WORLD, ierr);
  ierr = NavierStokesComputeFlowField(ctx);                                         CHKERRABORT(PETSC_COMM_WORLD, ierr);
  ierr = NavierStokesContextDestroy(ctx);                                           CHKERRABORT(PETSC_COMM_WORLD, ierr);

  CHKMEMQ;
  PetscFinalize();
  PetscFunctionReturn(0);
}

#undef __FUNCT__  
#define __FUNCT__ "NavierStokesInitializePackage"
/*@C
  NavierStokesInitializePackage - This function initializes everything in the NavierStokes package.

  Input Parameter:
. path - The dynamic library path, or PETSC_NULL

  Level: developer

.keywords: NavierStokes, initialize, package
.seealso: PetscInitialize()
@*/
int NavierStokesInitializePackage(char *path) {
  static PetscTruth initialized = PETSC_FALSE;
  char              logList[256];
  char             *className;
  PetscTruth        opt;
  int               ierr;

  PetscFunctionBegin;
  if (initialized == PETSC_TRUE) PetscFunctionReturn(0);
  initialized = PETSC_TRUE;
  /* Register Classes */
  ierr = PetscLogClassRegister(&NAVIERSTOKES_COOKIE, "NavierStokes");                                     CHKERRQ(ierr);
  /* Register Constructors and Serializers */
  /* Register Events */
  ierr = PetscLogEventRegister(&NAVIERSTOKES_ComputeField, "NavierStokesComputeField", NAVIERSTOKES_COOKIE); CHKERRQ(ierr);
  /* Process info exclusions */
  ierr = PetscOptionsGetString(PETSC_NULL, "-log_info_exclude", logList, 256, &opt);                      CHKERRQ(ierr);
  if (opt == PETSC_TRUE) {
    ierr = PetscStrstr(logList, "NavierStokesContext", &className);                                       CHKERRQ(ierr);
    if (className) {
      ierr = PetscLogInfoDeactivateClass(NAVIERSTOKES_COOKIE);                                            CHKERRQ(ierr);
    }
  }
  /* Process summary exclusions */
  ierr = PetscOptionsGetString(PETSC_NULL, "-log_summary_exclude", logList, 256, &opt);                   CHKERRQ(ierr);
  if (opt == PETSC_TRUE) {
    ierr = PetscStrstr(logList, "NavierStokesContext", &className);                                       CHKERRQ(ierr);
    if (className) {
      ierr = PetscLogEventDeactivateClass(NAVIERSTOKES_COOKIE);                                           CHKERRQ(ierr);
    }
  }
  PetscFunctionReturn(0);
}

/*----------------------------------------- NavierStokesContext Creation ---------------------------------------------*/
#undef  __FUNCT__
#define __FUNCT__ "NavierStokesContextCreate"
/*@
  NavierStokesContextCreate - This function initializes the NavierStokes context.

  Collective on MPI_Comm

  Input Parameter:
. comm - The communicator

  Output Parameter:
. sCtx - The NavierStokesContext

  Level: beginner

.keywords: NavierStokes, context, create
.seealso: NavierStokesContextDestroy(), NavierStokesContextPrint(), NavierStokesContextSetup()
@*/
int NavierStokesContextCreate(MPI_Comm comm, NavierStokesContext *sCtx) {
  NavierStokesContext ctx;

  PetscFunctionBegin;
  /* Setup context */
  PetscHeaderCreate(ctx, _NavierStokesContext, int, NAVIERSTOKES_COOKIE, -1, "context", comm, 0, 0);
  PetscLogObjectCreate(ctx);
  PetscLogObjectMemory(ctx, sizeof(struct _NavierStokesContext));

  /* Initialize subobjects */
  ctx->grid          = PETSC_NULL;
  ctx->snes          = PETSC_NULL;
  ctx->snesSecondary = PETSC_NULL;
  ctx->sles          = PETSC_NULL;
  ctx->A             = PETSC_NULL;
  ctx->u             = PETSC_NULL;
  ctx->p             = PETSC_NULL;
  ctx->f             = PETSC_NULL;
  ctx->origF         = PETSC_NULL;
  ctx->uExact        = PETSC_NULL;
  ctx->pExact        = PETSC_NULL;
  ctx->RofS          = -1;
  /* Setup domain */
  ctx->geometryCtx.size[0]  = 2.0;
  ctx->geometryCtx.size[1]  = 2.0;
  ctx->geometryCtx.start[0] = 0.0;
  ctx->geometryCtx.start[1] = 0.0;
  /* Setup refinement */
  ctx->geometryCtx.maxArea  = 0.5;
  ctx->geometryCtx.areaFunc = PointFunctionConstant;
  ctx->geometryCtx.areaCtx  = PETSC_NULL;
  /* Initialize physical information */
  ctx->physicalCtx.nu      = -1.0;
  ctx->physicalCtx.fluidSG = 1.0;
  /* Initialize preconditioning suites */
  ctx->useConvection = PETSC_FALSE;
  ctx->useML         = PETSC_FALSE;
  ctx->useSchur      = PETSC_FALSE;
  ctx->useSecondary  = PETSC_FALSE;
  /* Initialize problem loop */
  ctx->dim        = 2;
  ctx->linear     = PETSC_FALSE;
  ctx->refineStep = 0;
  ctx->numLoops   = 0;
  /* Initialize viewers */
  ctx->mathViewer         = PETSC_NULL;
  ctx->nonlinearViewer    = PETSC_NULL;
  ctx->nonlinearHist      = PETSC_NULL;
  ctx->secondaryViewer    = PETSC_NULL;
  ctx->secondaryHist      = PETSC_NULL;
  ctx->linearViewer       = PETSC_NULL;
  ctx->linearHist         = PETSC_NULL;
  ctx->totalLinearViewer  = PETSC_NULL;
  ctx->totalLinearHist    = PETSC_NULL;
  ctx->innerLinearViewer  = PETSC_NULL;
  ctx->innerLinearHist    = PETSC_NULL;
  ctx->schurLinearViewer  = PETSC_NULL;
  ctx->schurLinearHist    = PETSC_NULL;

  *sCtx = ctx;
  PetscFunctionReturn(0);
}

#undef  __FUNCT__
#define __FUNCT__ "NavierStokesContextDestroy"
/*@
  NavierStokesContextDestroy - This function destroys the NavierStokes context.

  Collective on NavierStokesContext

  Input Parameter:
. ctx - The NavierStokesContext

  Level: beginner

.keywords: NavierStokes, context, destroy
.seealso: NavierStokesContextCreate(), NavierStokesContextSetup()
@*/
int NavierStokesContextDestroy(NavierStokesContext ctx) {
  Grid grid = ctx->grid;
  int  ierr;

  PetscFunctionBegin;
  if (--ctx->refct > 0) SETERRQ(PETSC_ERR_PLIB, "NavierStokes context should not be referenced more than once");
  PetscLogStagePush(UserStages[Stage_Setup]);
  ierr = GridFinalizeBC(grid);                                                                            CHKERRQ(ierr);
  ierr = GridDestroy(grid);                                                                               CHKERRQ(ierr);
  /* Destroy viewers */
  if (ctx->mathViewer != PETSC_NULL) {
    ierr = PetscViewerDestroy(ctx->mathViewer);                                                           CHKERRQ(ierr);
  }
  /* Destroy histograms */
  ierr = NavierStokesResetHists(ctx, PETSC_TRUE);                                                         CHKERRQ(ierr);
  ierr = NavierStokesDestroyHists(ctx);                                                                   CHKERRQ(ierr);
  /* Destroy logging stages */
  ierr = PetscFree(UserStages);                                                                           CHKERRQ(ierr);
  PetscLogStagePop();

  PetscLogObjectDestroy(ctx);
  PetscHeaderDestroy(ctx);
  PetscFunctionReturn(0);
}

/*------------------------------------------- NavierStokesContext Setup ----------------------------------------------*/
#undef  __FUNCT__
#define __FUNCT__ "NavierStokesContextSetup"
/*@
  NavierStokesContextSetup - This function configures the context from user options.

  Collective on NavierStokesContext

  Input Parameter:
. ctx - A NavierStokesContext

  Options Database Keys:
+ dim <num>               - The problem dimension
. linear                  - Using a linear discretization
. num_systems <num>       - The number of systems to solve
. refine_step <step>      - The step at which to start refining the mesh
. use_convection          - Include convection in the momentum equations
. use_secondary           - Use a second solver after the nonlinear krylov method
. physical_nu <nu>        - Set the kinematic viscosity of the fluid
. physical_fluid_sg <num> - Set the specific gravity of the fluid
. mesh_max_area <num>     - The maximum area of a triangle in the refined mesh
. pc_ml                   - The ML preconditioning suite
. pc_schur                - The Schur preconditioning suite
- math_viewer             - Opens a connection to Mathematica

  Level: beginner

.seealso NavierStokesRefineGrid(), NavierStokesDestroyGrid()
@*/
int NavierStokesSetup(NavierStokesContext ctx) {
  char       stageName[256];
  int        loop;
  PetscTruth opt;
  int        ierr;

  PetscFunctionBegin;
  /* Determine the problem dimension */
  ierr = PetscOptionsGetInt(PETSC_NULL, "-dim", &ctx->dim, &opt);                                         CHKERRQ(ierr);
  /* Determine the element type */
  ierr = PetscOptionsHasName(PETSC_NULL, "-linear", &ctx->linear);                                        CHKERRQ(ierr);
  /* The first iteration at which to refine the mesh */
  ierr = PetscOptionsGetInt(PETSC_NULL, "-refine_step", &ctx->refineStep, &opt);                          CHKERRQ(ierr);
  /* Determine how many systems to solve */
  ierr = PetscOptionsGetInt(PETSC_NULL, "-num_systems", &ctx->numLoops, &opt);                            CHKERRQ(ierr);
  /* Setup refinement */
  ierr = PetscOptionsGetReal("mesh", "-max_area", &ctx->geometryCtx.maxArea, &opt);                       CHKERRQ(ierr);
  /* Get physical information */
  ierr = PetscOptionsGetReal("physical_", "-nu",       &ctx->physicalCtx.nu,      &opt);                  CHKERRQ(ierr);
  ierr = PetscOptionsGetReal("physical_", "-fluid_sg", &ctx->physicalCtx.fluidSG, &opt);                  CHKERRQ(ierr);
  /* Setup preconditioning suites */
  ierr = PetscOptionsHasName(PETSC_NULL, "-pc_ml",    &ctx->useML);                                       CHKERRQ(ierr);
  ierr = PetscOptionsHasName(PETSC_NULL, "-pc_schur", &ctx->useSchur);                                    CHKERRQ(ierr);
  /* Setup convection */
  ierr = PetscOptionsHasName(PETSC_NULL, "-use_convection", &ctx->useConvection);                         CHKERRQ(ierr);
  /* Setup solver */
  ierr = PetscOptionsHasName(PETSC_NULL, "-use_secondary", &ctx->useSecondary);                           CHKERRQ(ierr);
  /* Create a Mathematica viewer */
  ctx->mathViewer = PETSC_NULL;
  ierr = PetscOptionsHasName(PETSC_NULL, "-math_viewer", &opt);                                           CHKERRQ(ierr);
  if (opt) {
    ierr = PetscViewerMathematicaOpen(ctx->comm, PETSC_DECIDE, PETSC_NULL, PETSC_NULL, &ctx->mathViewer); CHKERRQ(ierr);
  }

  /* Create iteration histograms */
  ierr = NavierStokesCreateHists(ctx);                                                                    CHKERRQ(ierr);

  /* Register logging stages */
  ierr = PetscMalloc((ctx->numLoops+NumStages) * sizeof(int), &UserStages);                               CHKERRQ(ierr);
  ierr = PetscLogStageRegister(&UserStages[Stage_Setup],      "Setup/Cleanup");                           CHKERRQ(ierr);
  ierr = PetscLogStageRegister(&UserStages[Stage_GridRefine], "Grid Refinement");                         CHKERRQ(ierr);
  for(loop = 0; loop < ctx->numLoops; loop++) {
    sprintf(stageName, "Trial %d", loop);
    ierr = PetscLogStageRegister(&UserStages[NumStages+loop], stageName);                                 CHKERRQ(ierr);
  }

  /* Create main problem */
  ierr = NavierStokesCreateGrid(ctx);                                                                     CHKERRQ(ierr);
  PetscFunctionReturn(0);
}

/*----------------------------------------------- Grid Creation -----------------------------------------------------*/
#undef  __FUNCT__
#define __FUNCT__ "NavierStokesContextCreateMeshBoundary"
/*@
  NavierStokesContextCreateMeshBoundary - This function creates a mesh boundary for the main problem.

  Collective on StokesContext

  Input Parameter:
. ctx - The NavierStokesContext with problem specific information

  Level: beginner

.seealso NavierStokesContextDestroyMeshBoundary(), NavierStokesContextCreateMesh()
@*/
int NavierStokesContextCreateMeshBoundary(NavierStokesContext ctx) {
  MPI_Comm             comm;
  MeshGeometryContext *geomCtx = &ctx->geometryCtx;
  int                  ierr;

  PetscFunctionBegin;
  ierr = PetscObjectGetComm((PetscObject) ctx, &comm);                                                    CHKERRQ(ierr);
  switch(ctx->dim) {
  case 1:
    ierr = MeshBoundary1DCreateSimple(comm, geomCtx, &ctx->boundaryCtx);                                  CHKERRQ(ierr);
    break;
  case 2:
    ierr = MeshBoundary2DCreateSimple(comm, geomCtx, &ctx->boundaryCtx);                                  CHKERRQ(ierr);
    break;
  default:
    SETERRQ1(PETSC_ERR_SUP, "Do not support meshes of dimension %d", ctx->dim);
  }
  PetscFunctionReturn(0);
}

#undef  __FUNCT__
#define __FUNCT__ "NavierStokesContextDestroyMeshBoundary"
/*@
  NavierStokesContextDestroyMeshBoundary - This function destroys a mesh boundary for the main problem.

  Collective on NavierStokesContext

  Input Parameter:
. ctx - The NavierStokesContext with problem specific information

  Level: beginner

.seealso NavierStokesContextCreateMeshBoundary(), NavierStokesContextDestroyMesh()
@*/
int NavierStokesContextDestroyMeshBoundary(NavierStokesContext ctx) {
  MPI_Comm comm;
  int      ierr;

  PetscFunctionBegin;
  ierr = PetscObjectGetComm((PetscObject) ctx, &comm);                                                    CHKERRQ(ierr);
  switch(ctx->dim) {
  case 1:
    ierr = MeshBoundary1DDestroy(comm, &ctx->boundaryCtx);                                                CHKERRQ(ierr);
    break;
  case 2:
    ierr = MeshBoundary2DDestroy(comm, &ctx->boundaryCtx);                                                CHKERRQ(ierr);
    break;
  default:
    SETERRQ1(PETSC_ERR_SUP, "Do not support meshes of dimension %d", ctx->dim);
  }
  PetscFunctionReturn(0);
}

#undef  __FUNCT__
#define __FUNCT__ "NavierStokesContextCreateMesh"
/*@
  NavierStokesContextCreateMesh - This function creates a mesh for the main problem.

  Collective on NavierStokesContext

  Input Parameter:
. ctx - A NavierStokesContext with problem specific information

  Output Parameter:
. m   - The Mesh

  Options Database Keys:
. mesh_refine   - Refines the mesh based on area criteria
. mesh_max_area - The maximum area of an element

  Level: beginner

.seealso NavierStokesRefineGrid(), NavierStokesDestroyGrid()
@*/
int NavierStokesContextCreateMesh(NavierStokesContext ctx, Mesh *m) {
  MeshGeometryContext *geomCtx = &ctx->geometryCtx;
  MPI_Comm             comm;
  Mesh                 mesh;
  Partition            part;
  int                  totalElements, totalNodes, totalEdges;
  char                 name[1024];
  int                  d;
  int                  ierr;

  PetscFunctionBegin;
  ierr = PetscObjectGetComm((PetscObject) ctx, &comm);                                                    CHKERRQ(ierr);
  ierr = NavierStokesContextCreateMeshBoundary(ctx);                                                      CHKERRQ(ierr);
  ierr = MeshCreate(comm, &mesh);                                                                         CHKERRQ(ierr);
  ierr = MeshSetDimension(mesh, ctx->dim);                                                                CHKERRQ(ierr);
  for(d = 0; d < ctx->dim; d++) {
    ierr = MeshSetPeriodicDimension(mesh, d, geomCtx->isPeriodic[d]);                                     CHKERRQ(ierr);
  }
  switch(ctx->dim) {
  case 1:
    if (ctx->linear == PETSC_TRUE) {
      ierr = MeshSetNumCorners(mesh, 2);                                                                  CHKERRQ(ierr);
    } else {
      ierr = MeshSetNumCorners(mesh, 3);                                                                  CHKERRQ(ierr);
    }
    break;
  case 2:
    if (ctx->linear == PETSC_TRUE) {
      ierr = MeshSetNumCorners(mesh, 3);                                                                  CHKERRQ(ierr);
    } else {
      ierr = MeshSetNumCorners(mesh, 6);                                                                  CHKERRQ(ierr);
    }
    break;
  default:
    SETERRQ1(PETSC_ERR_SUP, "Do not support meshes of dimension %d", ctx->dim);
  }
  ierr = MeshSetBoundary(mesh, &ctx->boundaryCtx);                                                        CHKERRQ(ierr);
  sprintf(name, "mesh.r%.4g", geomCtx->maxArea);
  ierr = PetscObjectSetName((PetscObject) mesh, name);                                                    CHKERRQ(ierr);
  ierr = MeshSetFromOptions(mesh);                                                                        CHKERRQ(ierr);
  ierr = NavierStokesContextDestroyMeshBoundary(ctx);                                                     CHKERRQ(ierr);
  /* Setup refinement */
  if (ctx->geometryCtx.areaCtx != PETSC_NULL) {
    ierr = MeshSetUserContext(mesh, geomCtx->areaCtx);                                                    CHKERRQ(ierr);
  } else {
    ierr = MeshSetUserContext(mesh, &geomCtx->maxArea);                                                   CHKERRQ(ierr);
  }
  /* Report on mesh */
  ierr = MeshGetPartition(mesh, &part);                                                                   CHKERRQ(ierr);
  switch(ctx->dim) {
  case 1:
    ierr = PartitionGetTotalElements(part, &totalElements);                                               CHKERRQ(ierr);
    ierr = PartitionGetTotalNodes(part, &totalNodes);                                                     CHKERRQ(ierr);
    PetscPrintf(ctx->comm, "Elements: %d Nodes: %d\n", totalElements, totalNodes);
    break;
  case 2:
    ierr = PartitionGetTotalElements(part, &totalElements);                                               CHKERRQ(ierr);
    ierr = PartitionGetTotalNodes(part, &totalNodes);                                                     CHKERRQ(ierr);
    ierr = PartitionGetTotalEdges(part, &totalEdges);                                                     CHKERRQ(ierr);
    PetscPrintf(ctx->comm, "Elements: %d Nodes: %d Edges: %d\n", totalElements, totalNodes, totalEdges);
    break;
  default:
    SETERRQ1(PETSC_ERR_SUP, "Do not support meshes of dimension %d", ctx->dim);
  }

  *m = mesh;
  PetscFunctionReturn(0);
}

#undef  __FUNCT__
#define __FUNCT__ "NavierStokesCreateGrid"
/*@
  NavierStokesCreateGrid - This function creates a grid for the main problem.

  Collective on NavierStokesContext

  Input Parameter:
. ctx     - A NavierStokesContext with problem specific information

  Level: beginner

.seealso NavierStokesRefineGrid(), NavierStokesDestroyGrid()
@*/
int NavierStokesCreateGrid(NavierStokesContext ctx) {
  Mesh mesh;
  Grid grid;
  int  ierr;

  PetscFunctionBegin;
  /* Construct the mesh */
  ierr = NavierStokesContextCreateMesh(ctx, &mesh);                                                       CHKERRQ(ierr);
  /* Construct the grid */
  ierr = GridCreate(mesh, &grid);                                                                         CHKERRQ(ierr);
  ierr = MeshDestroy(mesh);                                                                               CHKERRQ(ierr);

  switch(ctx->dim) {
  case 1:
    if (ctx->linear == PETSC_TRUE) {
      ierr = GridAddField(grid, "Velocity", DISCRETIZATION_TRIANGULAR_1D_LINEAR,   2, PETSC_NULL);        CHKERRQ(ierr);
      ierr = GridAddField(grid, "Pressure", DISCRETIZATION_TRIANGULAR_1D_CONSTANT, 1, PETSC_NULL);        CHKERRQ(ierr);
    } else {
      ierr = GridAddField(grid, "Velocity", DISCRETIZATION_TRIANGULAR_1D_QUADRATIC, 2, PETSC_NULL);       CHKERRQ(ierr);
      ierr = GridAddField(grid, "Pressure", DISCRETIZATION_TRIANGULAR_1D_LINEAR,    1, PETSC_NULL);       CHKERRQ(ierr);
    }
    break;
  case 2:
    if (ctx->linear == PETSC_TRUE) {
      ierr = GridAddField(grid, "Velocity", DISCRETIZATION_TRIANGULAR_2D_LINEAR,   2, PETSC_NULL);        CHKERRQ(ierr);
      ierr = GridAddField(grid, "Pressure", DISCRETIZATION_TRIANGULAR_2D_CONSTANT, 1, PETSC_NULL);        CHKERRQ(ierr);
    } else {
      ierr = GridAddField(grid, "Velocity", DISCRETIZATION_TRIANGULAR_2D_QUADRATIC, 2, PETSC_NULL);       CHKERRQ(ierr);
      ierr = GridAddField(grid, "Pressure", DISCRETIZATION_TRIANGULAR_2D_LINEAR,    1, PETSC_NULL);       CHKERRQ(ierr);
    }
    break;
  default:
    SETERRQ1(PETSC_ERR_SUP, "Do not support meshes of dimension %d", ctx->dim);
  }
  ierr = GridSetFromOptions(grid);                                                                        CHKERRQ(ierr);

  ctx->grid = grid;
  PetscFunctionReturn(0);
}

#undef  __FUNCT__
#define __FUNCT__ "NavierStokesRefineGrid"
/*@
  NavierStokesRefineGrid - This function refines the mesh for the main grid.

  Collective on NavierStokesContext

  Input Parameters:
+ ctx - A NavierStokesContext

  Options Database Keys:
. mesh_max_area - The maximum area an element may have

  Level: beginner

.seealso NavierStokesCreateGrid(), NavierStokesDestroyGrid()
@*/
int NavierStokesRefineGrid(NavierStokesContext ctx) {
  Grid                 oldGrid = ctx->grid;
  Grid                 grid;
  Mesh                 mesh;
  Partition            part;
  MeshGeometryContext *geomCtx = &ctx->geometryCtx;
  char                 name[1024];
  int                  totalElements, totalNodes, totalEdges;
  int                  ierr;

  PetscFunctionBegin;
  /* Construct a refined mesh */
  if (geomCtx->areaCtx != PETSC_NULL) {
    ierr = GridRefineMesh(oldGrid, geomCtx->areaFunc, &grid);                                             CHKERRQ(ierr);
    ierr = GridGetMesh(grid, &mesh);                                                                      CHKERRQ(ierr);
    ierr = MeshSetUserContext(mesh, geomCtx->areaCtx);                                                    CHKERRQ(ierr);
  } else {
    geomCtx->maxArea *= 0.5;
    ierr = GridRefineMesh(oldGrid, geomCtx->areaFunc, &grid);                                             CHKERRQ(ierr);
    ierr = GridGetMesh(grid, &mesh);                                                                      CHKERRQ(ierr);
    ierr = MeshSetUserContext(mesh, &geomCtx->maxArea);                                                   CHKERRQ(ierr);
  }
  sprintf(name, "mesh.r%.4g", geomCtx->maxArea);
  ierr = PetscObjectSetName((PetscObject) mesh, name);                                                    CHKERRQ(ierr);
  ierr = MeshSetOptionsPrefix(mesh, "ref_");                                                              CHKERRQ(ierr);
  ierr = GridSetOptionsPrefix(grid, "ref_");                                                              CHKERRQ(ierr);
  ierr = GridSetFromOptions(grid);                                                                        CHKERRQ(ierr);

  ierr = MeshGetPartition(mesh, &part);                                                                   CHKERRQ(ierr);
  ierr = PartitionGetTotalElements(part, &totalElements);                                                 CHKERRQ(ierr);
  ierr = PartitionGetTotalNodes(part, &totalNodes);                                                       CHKERRQ(ierr);
  ierr = PartitionGetTotalEdges(part, &totalEdges);                                                       CHKERRQ(ierr);
  PetscPrintf(ctx->comm, "Elements: %d Nodes: %d Edges: %d\n", totalElements, totalNodes, totalEdges);
  CHKMEMQ;

  /* Replace old grid with refined grid */
  ierr = GridFinalizeBC(oldGrid);                                                                         CHKERRQ(ierr);
  ierr = GridDestroy(oldGrid);                                                                            CHKERRQ(ierr);
  ctx->grid = grid;
  PetscFunctionReturn(0);
}

/*------------------------------------------------ Grid Setup -------------------------------------------------------*/
#undef  __FUNCT__
#define __FUNCT__ "NavierStokesSetupGrid"
/*@
  NavierStokesSetupGrid - This function sets all the functions, operators, and boundary conditions for the problem.
  It also sets the parameters associated with the fields.

  Collective on Grid

  Input Parameters:
+ grid - The grid
- ctx  - A NavierStokesContext

  Options Database Keys:
. use_laplacian - Use the Laplacian instead of the Rate-of-Strain tensor

  Level: intermediate

.seealso NavierStokesCreateGrid()
@*/
int NavierStokesSetupGrid(NavierStokesContext ctx) {
  Grid       grid   = ctx->grid;
  int        vel    = 0;
  int        pres   = 1;
  double     nu     = ctx->physicalCtx.nu;
  double     invRho = 1.0/ctx->physicalCtx.fluidSG;
  int        LC;
  PetscTruth opt;
  int        ierr;

  PetscFunctionBegin;
  /* Setup Fields */
  ierr = GridSetFieldName(grid, vel,  "Velocity");                                                        CHKERRQ(ierr);
  ierr = GridSetFieldName(grid, pres, "Pressure");                                                        CHKERRQ(ierr);

  /* Setup Problem */
  ierr = GridSetActiveField(grid, vel);                                                                   CHKERRQ(ierr);
  if (ctx->useML == PETSC_FALSE) {
    ierr = GridAddActiveField(grid, pres);                                                                CHKERRQ(ierr);
  }

  /* Setup Rhs */
  ierr = PetscOptionsHasName(PETSC_NULL, "-use_laplacian", &opt);                                         CHKERRQ(ierr);
  if (opt) {
    ctx->RofS = LAPLACIAN;
  } else {
    ierr = GridRegisterOperator(grid, vel, RateOfStrainTensor, &ctx->RofS);                               CHKERRQ(ierr);
  }

  ierr = GridSetRhsOperator(grid,   vel,  vel,  ctx->RofS,  -nu, PETSC_FALSE, PETSC_NULL);                CHKERRQ(ierr);
  if (ctx->useML == PETSC_FALSE) {
    ierr = GridAddRhsOperator(grid, pres, vel,  GRADIENT,  -invRho, PETSC_FALSE, PETSC_NULL);             CHKERRQ(ierr);
    ierr = GridAddRhsOperator(grid, vel,  pres, DIVERGENCE, invRho, PETSC_FALSE, PETSC_NULL);             CHKERRQ(ierr);
  }
  if (ctx->useConvection == PETSC_TRUE) {
    ierr = GridAddRhsNonlinearOperator(grid, vel, ConvectionOperator, 1.0, PETSC_FALSE);                  CHKERRQ(ierr);
  }
  ierr = NavierStokesSetupRhsFunction(grid, ctx);                                                         CHKERRQ(ierr);

  /* Setup Jacobian */
  ierr = GridSetMatOperator(grid,   vel,  vel,  ctx->RofS,   -nu, PETSC_FALSE, PETSC_NULL);               CHKERRQ(ierr);
  if (ctx->useML == PETSC_FALSE) {
    ierr = GridAddMatOperator(grid, pres, vel,  GRADIENT,   -invRho, PETSC_FALSE, PETSC_NULL);            CHKERRQ(ierr);
    ierr = GridAddMatOperator(grid, vel,  pres, DIVERGENCE,  invRho, PETSC_FALSE, PETSC_NULL);            CHKERRQ(ierr);
  }
  if (ctx->useConvection == PETSC_TRUE) {
    ierr = GridRegisterOperator(grid, vel, LinearConvectionOperator, &LC);                                CHKERRQ(ierr);
    ierr = GridAddMatOperator(grid, vel, vel, LC, 1.0, PETSC_FALSE, PETSC_NULL);                          CHKERRQ(ierr);
  }

  /* Setup Dirchlet boundary conditions */
  ierr = NavierStokesSetupBC(grid, ctx);                                                                  CHKERRQ(ierr);
  PetscFunctionReturn(0);
}

#undef  __FUNCT__
#define __FUNCT__ "NavierStokesSetupRhsFunction"
/*@
  NavierStokesSetupRhsFunction - This function chooses a forcing function for the problem.

  Collective on Grid

  Input Parameters:
+ grid - The grid
- ctx  - A NavierStokesContext

  Level: intermediate

.seealso NavierStokesSetupGrid()
@*/
int NavierStokesSetupRhsFunction(Grid grid, NavierStokesContext ctx)
{
  double alpha = 1.0;
  int    vel  = 0;
  int    pres = 1;
  int    ierr;

  PetscFunctionBegin;
  if (ctx->useConvection == PETSC_TRUE) alpha = -1.0;
  ierr = GridAddRhsFunction(grid, vel, VelocityRhsFunction, alpha);                                       CHKERRQ(ierr);
  if (ctx->useML == PETSC_FALSE) {
    ierr = GridAddRhsFunction(grid, pres, PointFunctionZero, 1.0);                                        CHKERRQ(ierr);
  }
  PetscFunctionReturn(0);
}

#undef  __FUNCT__
#define __FUNCT__ "NavierStokesSetupBC"
/*@
  NavierStokesSetupBC - This function chooses boundary conditions for the problem.

  Collective on Grid

  Input Parameters:
+ grid - The grid
- ctx  - A NavierStokesContext

  Level: intermediate

  Options Database Keys:
+ bc_reduce         - Explicitly reduce the system using boundary conditions
. bc_exact          - Use the exact solution for boundary conditions
. bc_pressure_none  - Do not impose boundary conditions on the pressure
. bc_pressure_full  - Impose boundary conditions over the entire boundary
. bc_pressure_exact - Use the exact solution for boundary conditions
. bc_pressure_x     - The x-coordinate of the pressure normalization point
. bc_pressure_y     - The y-coordinate of the pressure normalization point
- bc_pressure_z     - The z-coordinate of the pressure normalization point

.seealso NavierStokesSetupGrid()
@*/
int NavierStokesSetupBC(Grid grid, NavierStokesContext ctx)
{
  MeshGeometryContext *geomCtx = &ctx->geometryCtx;
  int                  vel     = 0;
  int                  pres    = 1;
  PetscTruth           reduceSystem;
  double               pX, pY, pZ;
  PetscTruth           opt;
  int                  ierr;

  PetscFunctionBegin;
  /* Setup reduction usng boundary conditions */
  ierr = PetscOptionsHasName(PETSC_NULL, "-bc_reduce", &reduceSystem);                                    CHKERRQ(ierr);

  ierr = PetscOptionsHasName(PETSC_NULL, "-bc_exact", &opt);                                              CHKERRQ(ierr);
  if (opt == PETSC_TRUE) {
    ierr = GridSetBC(grid, OUTER_BD, vel, VelocitySolutionFunction, reduceSystem);                        CHKERRQ(ierr);
  } else {
    ierr = GridSetBC(grid, OUTER_BD, vel, PointFunctionZero, reduceSystem);                               CHKERRQ(ierr);
  }

  /* Pressure boundary conditions */
  ierr = PetscOptionsHasName(PETSC_NULL, "-bc_pressure_none", &opt);                                      CHKERRQ(ierr);
  if (opt == PETSC_FALSE) {
    ierr = PetscOptionsHasName(PETSC_NULL, "-bc_pressure_full", &opt);                                    CHKERRQ(ierr);
    if (opt == PETSC_TRUE) {
      ierr = GridAddBC(grid, OUTER_BD, pres, PressureSolutionFunction, reduceSystem);                     CHKERRQ(ierr);
    } else {
      pX   = geomCtx->start[0];
      pY   = geomCtx->start[1];
      pZ   = geomCtx->start[2];
      ierr = PetscOptionsGetReal(PETSC_NULL, "-bc_pressure_x", &pX, &opt);                                CHKERRQ(ierr);
      ierr = PetscOptionsGetReal(PETSC_NULL, "-bc_pressure_y", &pY, &opt);                                CHKERRQ(ierr);
      ierr = PetscOptionsGetReal(PETSC_NULL, "-bc_pressure_z", &pZ, &opt);                                CHKERRQ(ierr);
      ierr = PetscOptionsHasName(PETSC_NULL, "-bc_pressure_exact", &opt);                                 CHKERRQ(ierr);
      if (opt == PETSC_TRUE) {
        ierr = GridSetPointBC(grid, pX, pY, pZ, pres, PressureSolutionFunction, reduceSystem);            CHKERRQ(ierr);
      } else {
        ierr = GridSetPointBC(grid, pX, pY, pZ, pres, PointFunctionZero, reduceSystem);                   CHKERRQ(ierr);
      }
    }
  }
  ierr = GridSetBCContext(grid, ctx);                                                                     CHKERRQ(ierr);
  PetscFunctionReturn(0);
}

/*------------------------------------------------ SNES Setup -------------------------------------------------------*/
#undef  __FUNCT__
#define __FUNCT__ "NavierStokesCreateStructures"
/*@
  NavierStokesCreateStructures - Create the vector and matrix storgage, and solvers.

  Collective on NavierStokesContext

  Input Parameter:
. ctx - A NavierStokesContext

  Level: beginner

.seealso NavierStokesRefineGrid(), NavierStokesDestroyGrid()
@*/
int NavierStokesCreateStructures(NavierStokesContext ctx) {
  VarOrdering pOrder;
  int         pres = 1;
  int         size;
  int         ierr;

  PetscFunctionBegin;
  if (ctx->useConvection) {
    /* Create the nonlinear solver */
    ierr = GSNESCreate(ctx->grid, ctx, &ctx->snes);                                                       CHKERRQ(ierr);
    ierr = SNESSetFromOptions(ctx->snes);                                                                 CHKERRQ(ierr);
    if (ctx->useSecondary == PETSC_TRUE) {
      ierr = GSNESCreate(ctx->grid, ctx, &ctx->snesSecondary);                                            CHKERRQ(ierr);
      ierr = SNESSetOptionsPrefix(ctx->snesSecondary, "sec_");                                            CHKERRQ(ierr);
      ierr = SNESSetFromOptions(ctx->snesSecondary);                                                      CHKERRQ(ierr);
    }
  } else {
    /* Create the linear solver */
    ierr = SLESCreate(ctx->comm, &ctx->sles);                                                             CHKERRQ(ierr);
    ierr = SLESSetFromOptions(ctx->sles);                                                                 CHKERRQ(ierr);
    /* Create the system matrix */
    ierr = GMatCreate(ctx->grid, &ctx->A);                                                                CHKERRQ(ierr);
    /* Create rhs and projected rhs vectors */
    ierr = GVecCreate(ctx->grid, &ctx->f);                                                                CHKERRQ(ierr);
    ierr = PetscObjectSetName((PetscObject) ctx->f, "Rhs");                                               CHKERRQ(ierr);
    ierr = GVecDuplicate(ctx->f, &ctx->origF);                                                            CHKERRQ(ierr);
    ierr = PetscObjectSetName((PetscObject) ctx->origF, "OriginalRhs");                                   CHKERRQ(ierr);
  }
  /* Create solution and exact solution vectors */
  ierr = GVecCreate(ctx->grid, &ctx->u);                                                                  CHKERRQ(ierr);
  ierr = PetscObjectSetName((PetscObject) ctx->u, "Solution");                                            CHKERRQ(ierr);
  ierr = GVecDuplicate(ctx->u, &ctx->uExact);                                                             CHKERRQ(ierr);
  ierr = PetscObjectSetName((PetscObject) ctx->uExact, "ExactSolution");                                  CHKERRQ(ierr);
  ierr = VecGetSize(ctx->u, &size);                                                                       CHKERRQ(ierr);
  PetscPrintf(ctx->comm, "Variables: %d\n", size);
  if (ctx->useML == PETSC_TRUE) {
    FieldClassMap cm, reduceCM;
    VarOrdering   reduceOrder;
    Vec           reduceVec;

    /* Create the multiplier, and the exact multiplier */
    ierr = VarOrderingCreateGeneral(ctx->grid, 1, &pres, &pOrder);                                        CHKERRQ(ierr);
    ierr = GVecCreateRectangular(ctx->grid, pOrder, &ctx->p);                                             CHKERRQ(ierr);
    ierr = PetscObjectSetName((PetscObject) ctx->p, "Pressure");                                          CHKERRQ(ierr);
    ierr = GVecDuplicate(ctx->p, &ctx->pExact);                                                           CHKERRQ(ierr);
    ierr = PetscObjectSetName((PetscObject) ctx->pExact, "ExactPressure");                                CHKERRQ(ierr);

    ierr = VarOrderingGetClassMap(pOrder, &cm);                                                           CHKERRQ(ierr);
    ierr = FieldClassMapReduce(cm, ctx->grid, &reduceCM);                                                 CHKERRQ(ierr);
    ierr = VarOrderingCreateReduceGeneral(ctx->grid, cm, reduceCM, &reduceOrder);                         CHKERRQ(ierr);
    ierr = GVecCreateRectangularGhost(ctx->grid, reduceOrder, &reduceVec);                                CHKERRQ(ierr);
    ierr = GridCalcBCValues_Private(ctx->grid, reduceOrder, reduceVec, PETSC_FALSE, PETSC_NULL);          CHKERRQ(ierr);
    ierr = PetscObjectCompose((PetscObject) ctx->p,      "reductionOrder", (PetscObject) reduceOrder);    CHKERRQ(ierr);
    ierr = PetscObjectCompose((PetscObject) ctx->pExact, "reductionOrder", (PetscObject) reduceOrder);    CHKERRQ(ierr);
    ierr = PetscObjectCompose((PetscObject) ctx->p,      "reductionVec",   (PetscObject) reduceVec);      CHKERRQ(ierr);
    ierr = PetscObjectCompose((PetscObject) ctx->pExact, "reductionVec",   (PetscObject) reduceVec);      CHKERRQ(ierr);
    ierr = FieldClassMapDestroy(reduceCM);                                                                CHKERRQ(ierr);
    ierr = VarOrderingDestroy(pOrder);                                                                    CHKERRQ(ierr);
    ierr = VarOrderingDestroy(reduceOrder);                                                               CHKERRQ(ierr);
    ierr = VecDestroy(reduceVec);                                                                         CHKERRQ(ierr);
  }
  PetscFunctionReturn(0);
}

#undef  __FUNCT__
#define __FUNCT__ "NavierStokesDestroySNESMonitor"
int NavierStokesDestroySNESMonitor(void *ctx) {
  GSNESErrorMonitorCtx *monCtx = (GSNESErrorMonitorCtx *) ctx;
  int                   ierr;

  PetscFunctionBegin;
  ierr = PetscViewerDestroy(monCtx->error_viewer);                                                        CHKERRQ(ierr);
  PetscFunctionReturn(0);
}

#undef  __FUNCT__
#define __FUNCT__ "NavierStokesSetupSNES"
int NavierStokesSetupSNES(SNES snes, NavierStokesContext ctx)
{
  GSNESErrorMonitorCtx *monCtx = &ctx->snesMonitorCtx;
  PetscViewer           v;
  PetscDraw             draw;
  char                  viewerType[256];
  PetscTruth            opt;
  int                   ierr;

  PetscFunctionBegin;
  if (ctx->useML) {
    ierr = SNESSetUpdate(ctx->snes, PCMultiLevelUpdateSNES);                                              CHKERRQ(ierr);
    ierr = SNESSetLineSearch(ctx->snes, PCMultiLevelCubicLineSearch, ctx);                                CHKERRQ(ierr);
  }
  /* Setup update function */
  ierr = SNESSetUpdate(snes, NavierStokesUpdateGSNES);                                                    CHKERRQ(ierr);
  /* Setup convergence monitors */
  ierr = PetscOptionsGetString(PETSC_NULL, "-error_viewer", viewerType, 255, &opt);                       CHKERRQ(ierr);
  if (opt) {
    ierr = PetscStrcasecmp(viewerType, "draw", &opt);                                                     CHKERRQ(ierr);
    if (opt == PETSC_TRUE) {
      v    = PETSC_VIEWER_DRAW_(ctx->comm);
      ierr = PetscViewerSetFormat(v, PETSC_VIEWER_DRAW_LG);                                               CHKERRQ(ierr);
      ierr = PetscViewerDrawGetDraw(v, 0, &draw);                                                         CHKERRQ(ierr);
      ierr = PetscDrawSetTitle(draw, "Error");                                                            CHKERRQ(ierr);
    } else {
      ierr = PetscStrcasecmp(viewerType, "text", &opt);                                                   CHKERRQ(ierr);
      if (opt == PETSC_TRUE) {
        v    = PETSC_VIEWER_STDOUT_(ctx->comm);
        ierr = PetscObjectReference((PetscObject) v);                                                     CHKERRQ(ierr);
      } else {
        SETERRQ1(PETSC_ERR_ARG_WRONG, "Invalid error viewer: %s", viewerType);
      }
    }
  } else {
    ierr = PetscViewerASCIIOpen(ctx->comm, "/dev/null", &v);                                              CHKERRQ(ierr);
  }
  monCtx->error_viewer      = v;
  monCtx->solutionFunc      = NavierStokesSolutionFunction;
  monCtx->ctx               = ctx;
  monCtx->solution          = ctx->uExact;
  monCtx->norm_error_viewer = PETSC_VIEWER_STDOUT_(ctx->comm);
  ierr = SNESSetMonitor(snes, GSNESErrorMonitor, monCtx, NavierStokesDestroySNESMonitor);                 CHKERRQ(ierr);
  PetscFunctionReturn(0);
}

#undef  __FUNCT__
#define __FUNCT__ "NavierStokesDestroyKSPMonitor"
int NavierStokesDestroyKSPMonitor(void *ctx) {
  GVecErrorKSPMonitorCtx *monCtx = (GVecErrorKSPMonitorCtx *) ctx;
  int                     ierr;

  PetscFunctionBegin;
  ierr = PetscViewerDestroy(monCtx->error_viewer);                                                        CHKERRQ(ierr);
  PetscFunctionReturn(0);
}

#undef  __FUNCT__
#define __FUNCT__ "NavierStokesSetupKSP"
int NavierStokesSetupKSP(KSP ksp, NavierStokesContext ctx) {
  GVecErrorKSPMonitorCtx *monCtx = &ctx->kspMonitorCtx;
  PetscViewer             v;
  PetscDraw               draw;
  char                    viewerType[256];
  PetscTruth              opt;
  int                     ierr;

  PetscFunctionBegin;
  /* Setup the multilevel preconditioner */
  if (ctx->useML == PETSC_TRUE) {
    ierr = KSPSetPreconditionerSide(ksp, PC_SYMMETRIC);                                                   CHKERRQ(ierr);
  }
  /* Setup convergence monitors */
  ierr = PetscOptionsGetString(PETSC_NULL, "-error_viewer", viewerType, 255, &opt);                       CHKERRQ(ierr);
  if (opt) {
    ierr = PetscStrcasecmp(viewerType, "draw", &opt);                                                     CHKERRQ(ierr);
    if (opt == PETSC_TRUE) {
      v    = PETSC_VIEWER_DRAW_(ctx->comm);
      ierr = PetscViewerSetFormat(v, PETSC_VIEWER_DRAW_LG);                                               CHKERRQ(ierr);
      ierr = PetscViewerDrawGetDraw(v, 0, &draw);                                                         CHKERRQ(ierr);
      ierr = PetscDrawSetTitle(draw, "Error");                                                            CHKERRQ(ierr);
    } else {
      ierr = PetscStrcasecmp(viewerType, "text", &opt);                                                   CHKERRQ(ierr);
      if (opt == PETSC_TRUE) {
        v    = PETSC_VIEWER_STDOUT_(ctx->comm);
        ierr = PetscObjectReference((PetscObject) v);                                                     CHKERRQ(ierr);
      } else {
        SETERRQ1(PETSC_ERR_ARG_WRONG, "Invalid error viewer: %s", viewerType);
      }
    }
  } else {
    ierr = PetscViewerASCIIOpen(ctx->comm, "/dev/null", &v);                                              CHKERRQ(ierr);
  }
  monCtx->error_viewer      = v;
  monCtx->solution          = ctx->uExact;
  monCtx->norm_error_viewer = PETSC_VIEWER_STDOUT_(ctx->comm);
  monCtx->isInner           = ctx->useConvection;
  ierr = KSPSetMonitor(ksp, GVecErrorKSPMonitor, monCtx, NavierStokesDestroyKSPMonitor);                  CHKERRQ(ierr);
  PetscFunctionReturn(0);
}


#undef  __FUNCT__
#define __FUNCT__ "NavierStokesSetupPC"
int NavierStokesSetupPC(PC pc, NavierStokesContext ctx)
{
#if 0
#ifdef PETSC_USE_DYNAMIC_LIBRARIES
  char        libName[1024];
  char        locLibName[1024];
  const char *PMDir;
  PetscTruth  found;
#endif
#endif
  int         vel  = 0;
  int         pres = 1;
  PetscTruth  opt;
  int         ierr;

  PetscFunctionBegin;
  /* Setup PM preconditioner library */
#if 0
#ifdef PETSC_USE_DYNAMIC_LIBRARIES
  ierr = PMGetPMDirectory(&PMDir);                                                                        CHKERRQ(ierr);
  ierr = PetscStrcpy(libName, PMDir);                                                                     CHKERRQ(ierr);
  ierr = PetscStrcat(libName, "/lib/lib$BOPT/$PETSC_ARCH/");                                              CHKERRQ(ierr);
  ierr = PetscStrcat(libName, "/libpmpc");                                                                CHKERRQ(ierr);
  ierr = PetscDLLibraryRetrieve(ctx->comm, libName, locLibName, 1024, &found);                            CHKERRQ(ierr);
  if (found == PETSC_FALSE) {
    PetscPrintf(ctx->comm, "Could not open PM preconditioner library %s\n", libName);
  } else {
    ierr = PetscDLLibraryAppend(ctx->comm, &DLLibrariesLoaded, libName);                                  CHKERRQ(ierr);
  }
#endif
#endif
  /* Setup the multilevel preconditioner */
  if (ctx->useML == PETSC_TRUE) {
#ifndef PETSC_USE_DYANMIC_LIBRARIES
    ierr = GSolverInitializePackage(PETSC_NULL);                                                          CHKERRQ(ierr);
#endif
    ierr = PetscOptionsHasName(PETSC_NULL, "-bc_reduce", &opt);                                           CHKERRQ(ierr);
    if (opt == PETSC_FALSE) SETERRQ(PETSC_ERR_ARG_WRONGSTATE, "The -bc_reduce option is necessary for ML");

    ierr = PCSetType(pc, PCMULTILEVEL);                                                                   CHKERRQ(ierr);
    ierr = PCMultiLevelSetFields(pc, pres, vel);                                                          CHKERRQ(ierr);
    ierr = PCMultiLevelSetGradientOperator(pc, GRADIENT, DIVERGENCE, -1.0);                               CHKERRQ(ierr);
    ierr = PCSetVector(pc, ctx->u);                                                                       CHKERRQ(ierr);
  }
  /* Setup the Schur complement preconditioner */
  if (ctx->useSchur == PETSC_TRUE) {
    ierr = PCSetType(pc, PCSCHUR);                                                                        CHKERRQ(ierr);
    ierr = PCSchurSetGradientOperator(pc, GRADIENT, DIVERGENCE);                                          CHKERRQ(ierr);
  }
  PetscFunctionReturn(0);
}

#undef  __FUNCT__
#define __FUNCT__ "NavierStokesSetupStructures"
int NavierStokesSetupStructures(NavierStokesContext ctx) {
  SLES         sles;
  KSP          ksp;
  PC           pc;
  VarOrdering  pOrder;
  int          vel  = 0;
  int          pres = 1;
  MatStructure flag;
  PetscTruth   opt;
  int          ierr;

  PetscFunctionBegin;
  if (ctx->useConvection == PETSC_TRUE) {
    /* Setup the nonlinear solver */
    ierr = NavierStokesSetupSNES(ctx->snes, ctx);                                                         CHKERRQ(ierr);
    /* Setup the linear solver */
    ierr = SNESGetSLES(ctx->snes, &sles);                                                                 CHKERRQ(ierr);
    ierr = SLESGetKSP(sles, &ksp);                                                                        CHKERRQ(ierr);
    ierr = NavierStokesSetupKSP(ksp, ctx);                                                                CHKERRQ(ierr);
    ierr = SLESGetPC(sles, &pc);                                                                          CHKERRQ(ierr);
    ierr = NavierStokesSetupPC(pc, ctx);                                                                  CHKERRQ(ierr);
    if (ctx->useSecondary == PETSC_TRUE) {
      /* Setup the nonlinear solver */
      ierr = NavierStokesSetupSNES(ctx->snesSecondary, ctx);                                              CHKERRQ(ierr);
      /* Setup the linear solver */
      ierr = SNESGetSLES(ctx->snesSecondary, &sles);                                                      CHKERRQ(ierr);
      ierr = SLESGetKSP(sles, &ksp);                                                                      CHKERRQ(ierr);
      ierr = NavierStokesSetupKSP(ksp, ctx);                                                              CHKERRQ(ierr);
      ierr = SLESGetPC(sles, &pc);                                                                        CHKERRQ(ierr);
      ierr = NavierStokesSetupPC(pc, ctx);                                                                CHKERRQ(ierr);
    }
  } else {
    /* Setup the linear solver */
    ierr = SLESSetOperators(ctx->sles, ctx->A, ctx->A, SAME_NONZERO_PATTERN);                             CHKERRQ(ierr);
    ierr = SLESGetKSP(ctx->sles, &ksp);                                                                   CHKERRQ(ierr);
    ierr = NavierStokesSetupKSP(ksp, ctx);                                                                CHKERRQ(ierr);
    ierr = SLESGetPC(ctx->sles, &pc);                                                                     CHKERRQ(ierr);
    ierr = NavierStokesSetupPC(pc, ctx);                                                                  CHKERRQ(ierr);
    /* Evaluate the rhs */
    ierr = GridEvaluateRhs(ctx->grid, ctx->u, ctx->f, (PetscObject) ctx);                                 CHKERRQ(ierr);
    ierr = VecCopy(ctx->f, ctx->origF);                                                                   CHKERRQ(ierr);
    if (ctx->mathViewer != PETSC_NULL) {
      ierr = GVecView(ctx->f, ctx->mathViewer);                                                           CHKERRQ(ierr);
    }
    /* Evaluate the system matrix */
    flag = DIFFERENT_NONZERO_PATTERN;
    ierr = GridEvaluateSystemMatrix(ctx->grid, ctx->u, &ctx->A, &ctx->A, &flag, (PetscObject) ctx);       CHKERRQ(ierr);
    ierr = MatCheckSymmetry(ctx->A);                                                                      CHKERRQ(ierr);
    /* Apply Dirchlet boundary conditions */
    ierr = GMatSetBoundary(ctx->A, 1.0, ctx);                                                             CHKERRQ(ierr);
    ierr = GVecSetBoundary(ctx->f, ctx);                                                                  CHKERRQ(ierr);
    /* View structures */
    ierr = PetscOptionsHasName(PETSC_NULL, "-mat_view", &opt);                                            CHKERRQ(ierr);
    if (opt) {
      ierr = MatView(ctx->A, PETSC_VIEWER_STDOUT_(ctx->comm));                                            CHKERRQ(ierr);
      ierr = MatView(ctx->A, PETSC_VIEWER_DRAW_(ctx->comm));                                              CHKERRQ(ierr);
    }
    ierr = PetscOptionsHasName(PETSC_NULL, "-vec_view", &opt);                                            CHKERRQ(ierr);
    if (opt) {
      ierr = VecView(ctx->f, PETSC_VIEWER_STDOUT_(ctx->comm));                                            CHKERRQ(ierr);
    }
  }

  /* Evaluate the exact solution */
  ierr = GVecEvaluateFunction(ctx->uExact, 1, &vel, VelocitySolutionFunction, 1.0, ctx);                  CHKERRQ(ierr);
  if (ctx->useML == PETSC_FALSE) {
    ierr = VarOrderingCreateGeneral(ctx->grid, 1, &pres, &pOrder);                                        CHKERRQ(ierr);
    ierr = GVecEvaluateFunctionRectangular(ctx->pExact, pOrder, PressureSolutionFunction, 1.0, ctx);      CHKERRQ(ierr);
    ierr = VarOrderingDestroy(pOrder);                                                                    CHKERRQ(ierr);
  }
  if (ctx->mathViewer != PETSC_NULL) {
    ierr = GVecView(ctx->uExact, ctx->mathViewer);                                                        CHKERRQ(ierr);
  }
  /* Evaluate the exact multiplier */
  if (ctx->useML == PETSC_TRUE) {
    ierr = VarOrderingCreateGeneral(ctx->grid, 1, &pres, &pOrder);                                        CHKERRQ(ierr);
    ierr = GVecEvaluateFunctionRectangular(ctx->pExact, pOrder, PressureSolutionFunction, 1.0, ctx);      CHKERRQ(ierr);
    ierr = VarOrderingDestroy(pOrder);                                                                    CHKERRQ(ierr);
  }
  PetscFunctionReturn(0);
}

#undef  __FUNCT__
#define __FUNCT__ "NavierStokesDestroyStructures"
int NavierStokesDestroyStructures(NavierStokesContext ctx) {
  int ierr;

  PetscFunctionBegin;
  if (ctx->useConvection == PETSC_TRUE) {
    ierr = GSNESDestroy(ctx->snes);                                                                       CHKERRQ(ierr);
    if (ctx->useSecondary == PETSC_TRUE) {
      ierr = GSNESDestroy(ctx->snesSecondary);                                                            CHKERRQ(ierr);
    }
  } else {
    ierr = SLESDestroy(ctx->sles);                                                                        CHKERRQ(ierr);
    ierr = GMatDestroy(ctx->A);                                                                           CHKERRQ(ierr);
    ierr = GVecDestroy(ctx->f);                                                                           CHKERRQ(ierr);
    ierr = GVecDestroy(ctx->origF);                                                                       CHKERRQ(ierr);
  }
  ierr = GVecDestroy(ctx->u);                                                                             CHKERRQ(ierr);
  ierr = GVecDestroy(ctx->uExact);                                                                        CHKERRQ(ierr);
  if (ctx->useML == PETSC_TRUE) {
    ierr = GVecDestroy(ctx->p);                                                                           CHKERRQ(ierr);
    ierr = GVecDestroy(ctx->pExact);                                                                      CHKERRQ(ierr);
  }
  PetscFunctionReturn(0);
}

/*---------------------------------------------- Sanity Checks ------------------------------------------------------*/
#undef  __FUNCT__
#define __FUNCT__ "MatCheckSymmetry"
int MatCheckSymmetry(Mat A)
{
  Mat        trA;
  PetscTruth isSym;
  int        ierr;

  PetscFunctionBegin;
  ierr = MatTranspose(A, &trA);                                                                           CHKERRQ(ierr);
  ierr = MatEqual(A, trA, &isSym);                                                                        CHKERRQ(ierr);
  ierr = MatDestroy(trA);                                                                                 CHKERRQ(ierr);
  if (isSym == PETSC_FALSE) PetscFunctionReturn(1);
  PetscFunctionReturn(0);
}

#undef  __FUNCT__
#define __FUNCT__ "NavierStokesCheckSolution"
int NavierStokesCheckSolution(NavierStokesContext ctx, GVec u, GVec p, const char type[]) {
  SLES        sles;
  GVec        r, projR;
  PC          pc;
  PetscScalar minusOne = -1.0;
  PetscScalar norm;
  int         ierr;

  PetscFunctionBegin;
  ierr = GVecDuplicate(u, &r);                                                                            CHKERRQ(ierr);
  ierr = GVecDuplicate(u, &projR);                                                                        CHKERRQ(ierr);
  if (ctx->useConvection == PETSC_TRUE) {
    ierr = SNESGetSLES(ctx->snes, &sles);                                                                 CHKERRQ(ierr);
    ierr = SLESGetPC(sles, &pc);                                                                          CHKERRQ(ierr);
  } else {
    ierr = SLESGetPC(ctx->sles, &pc);                                                                     CHKERRQ(ierr);
  }

  /* Check the velocity solution */
  if (ctx->useConvection == PETSC_TRUE) {
    ierr = GSNESEvaluateRhs(ctx->snes, u, r, (PetscObject) ctx);                                          CHKERRQ(ierr);
    ierr = GSNESRhsBC(ctx->snes, r, ctx);                                                                 CHKERRQ(ierr);
  } else {
    /* A u^* */
    ierr = MatMult(ctx->A, u, r);                                                                         CHKERRQ(ierr);
    /* f - A u^* */
    ierr = VecAYPX(&minusOne, ctx->origF, r);                                                             CHKERRQ(ierr);
  }
  /* P^T_2 (f - A u^*) */
  if (ctx->useML == PETSC_TRUE) {
    ierr = PCApplySymmetricLeft(pc, r, projR);                                                            CHKERRQ(ierr);
    ierr = VecNorm(projR, NORM_2, &norm);                                                                 CHKERRQ(ierr);
    PetscPrintf(ctx->comm, "Residual of the %s velocity solution: %lf\n", type, norm);
  } else {
    ierr = VecNorm(r, NORM_2, &norm);                                                                     CHKERRQ(ierr);
    PetscPrintf(ctx->comm, "Residual of the %s solution: %lf\n", type, norm);
  }

  /* Check the multiplier */
  if (ctx->useML == PETSC_TRUE) {
    /* P^T_2 B p^* */
    ierr = PCMultiLevelApplyGradient(pc, p, r);                                                           CHKERRQ(ierr);
    ierr = PCApplySymmetricLeft(pc, r, projR);                                                            CHKERRQ(ierr);
    ierr = VecNorm(projR, NORM_2, &norm);                                                                 CHKERRQ(ierr);
    PetscPrintf(ctx->comm, "Residual of the %s gradient contribution: %lf\n", type, norm);
    if (ctx->useConvection == PETSC_FALSE) {
      /* f - A u^* - B p^* */
      ierr = MatMult(ctx->A, u, projR);                                                                   CHKERRQ(ierr);
      ierr = VecAYPX(&minusOne, ctx->origF, projR);                                                       CHKERRQ(ierr);
      ierr = VecAXPY(&minusOne, r, projR);                                                                CHKERRQ(ierr);
      ierr = VecNorm(projR, NORM_2, &norm);                                                               CHKERRQ(ierr);
      PetscPrintf(ctx->comm, "Residual of the full %s solution: %lf\n", type, norm);
    }
  }
 
  ierr = GVecDestroy(r);                                                                                  CHKERRQ(ierr);
  ierr = GVecDestroy(projR);                                                                              CHKERRQ(ierr);
  PetscFunctionReturn(0);
}

/*---------------------------------------------- Problem Callbacks --------------------------------------------------*/
#undef  __FUNCT__
#define __FUNCT__ "NavierStokesSolutionFunction"
/*@
  NavierStokesSolutionFunction - This function is the solution function for the problem.

  Not collective

  Input Parameters:
+ n      - The number of points
. comp   - The number of components
. x,y,z  - The points
. values - The output
- ctx    - A NavierStokesContext

  Level: beginner

  Note:
  The solution is u = x^2 \hat x - 2 x y \hat y, p = x + y - 2

.keywords solution
.seealso VelocityRhsFunction(), PressureSolutionFunction()
@*/
int NavierStokesSolutionFunction(Vec u, void *ctx)
{
  NavierStokesContext s    = (NavierStokesContext) ctx;
  int                 vel  = 0;
  int                 pres = 1;
  int                 ierr;

  PetscFunctionBegin;
  ierr = GVecEvaluateFunction(u, 1, &vel, VelocitySolutionFunction, 1.0, ctx);                            CHKERRQ(ierr);
  if (s->useML == PETSC_FALSE) {
    ierr = GVecEvaluateFunction(u, 1, &pres, PressureSolutionFunction, 1.0, ctx);                         CHKERRQ(ierr);
  }
  PetscFunctionReturn(0);
}

#undef  __FUNCT__
#define __FUNCT__ "VelocitySolutionFunction"
/*@
  VelocitySolutionFunction - This function is the velocity solution function for the problem.

  Not collective

  Input Parameters:
+ n      - The number of points
. comp   - The number of components
. x,y,z  - The points
. values - The output
- ctx    - A NavierStokesContext

  Level: beginner

  Note:
  The solution is u = x^2 \hat x - 2 x y \hat y

.keywords velocity, solution
.seealso PressureSolutionFunction(), VelocityRhsFunction()
@*/
int VelocitySolutionFunction(int n, int comp, double *x, double *y, double *z, PetscScalar *values, void *ctx)
{
  int i;

  PetscFunctionBegin;
  if (comp != 2) SETERRQ1(PETSC_ERR_ARG_WRONG, "Invalid number of components %d", comp);
  for(i = 0; i < n; i++) {
    values[i*2+0] = x[i]*x[i];
    values[i*2+1] = -2.0*x[i]*y[i];
  }
  PetscFunctionReturn(0);
}

#undef  __FUNCT__
#define __FUNCT__ "PressureSolutionFunction"
/*@
  PressureSolutionFunction - This function is the pressure solution function for the problem.

  Not collective

  Input Parameters:
+ n      - The number of points
. comp   - The number of components
. x,y,z  - The points
. values - The output
- ctx    - A NavierStokesContext

  Level: beginner

  Note:
  The solution is p = x + y - 2, the integral of p is zero over the domain.

.keywords pressure, solution
.seealso VelocitySolutionFunction(), VelocityRhsFunction()
@*/
int PressureSolutionFunction(int n, int comp, double *x, double *y, double *z, PetscScalar *values, void *ctx)
{
  int i;

  PetscFunctionBegin;
  if (comp != 1) SETERRQ1(PETSC_ERR_ARG_WRONG, "Invalid number of components %d", comp);
  for(i = 0; i < n; i++) {
      values[i] = x[i] + y[i] - 2.0;
  }
  PetscFunctionReturn(0);
}

#undef  __FUNCT__
#define __FUNCT__ "VelocityRhsFunction"
/*@
  VelocityRhsFunction - This function is the forcing function for the problem.

  Not collective

  Input Parameters:
+ n      - The number of points
. comp   - The number of components
. x,y,z  - The points
. values - The output
- ctx    - A NavierStokesContext

  Level: beginner

  Note:
$ The rhs is
$     u \cdot \nabla u - \nu \Delta u - 1/\rho \nabla p = (2 x^3 - 2 \nu - 1/\rho) \hat x + (2 x^2 y - 1/\rho) \hat y
$ or
$     -\nu \Delta u - 1/\rho \nabla p = -(2\nu + 1/\rho) \hat x - \hat y
$ in the linear case.

.keywords velocity, rhs
.seealso VelocitySolutionFunction(), PressureSolutionFunction()
@*/
int VelocityRhsFunction(int n, int comp, double *x, double *y, double *z, PetscScalar *values, void *ctx) {
  NavierStokesContext s      = (NavierStokesContext) ctx;
  double              nu     = s->physicalCtx.nu;
  double              invRho = 1.0/s->physicalCtx.fluidSG;
  int                 i;

  PetscFunctionBegin;
  if (comp != 2) SETERRQ1(PETSC_ERR_ARG_WRONG, "Invalid number of components %d", comp);
  if (s->useConvection == PETSC_TRUE) {
    for (i = 0; i < n; i++) {
      values[i*2+0] = 2.0*x[i]*x[i]*x[i] - 2.0*nu - invRho;
      values[i*2+1] = 2.0*x[i]*x[i]*y[i] - invRho;
    }
  } else {
    for (i = 0; i < n; i++) {
      values[i*2+0] = -2.0*nu - invRho;
      values[i*2+1] = -invRho;
    }
  }
  PetscFunctionReturn(0);
}

/*--------------------------------------------- Analysis Functions --------------------------------------------------*/
#undef  __FUNCT__
#define __FUNCT__ "NavierStokesCreateHists"
/*@
  NavierStokesCreateHists - This function creates problem histograms.

  Collective on NavierStokesContext

  Input Parameter:
. ctx   - The NavierStokesContext

  Options Database Keys:
+ linear_viewer (null)          - Histogram linear iterations
. linear_hist_bins <bins>       - The number of histogram bins
. inner_linear_viewer (null)    - Histogram linear iterations in the preconditioner
. inner_linear_hist_bins <bins> - The number of histogram bins
. schur_linear_hist_bins <bins> - The number of histogram bins
. nonlinear_viewer (null)       - Histogram nonlinear iterations
- nonlinear_hist_bins <bins>    - The number of histogram bins

  Level: advanced

.seealso NavierStokesResetHists(), NavierStokesDestroyHists()
@*/
int NavierStokesCreateHists(NavierStokesContext ctx)
{
  PetscDraw   draw;
  char        viewerType[256];
  int         bins;
  PetscTruth  opt;
  int         ierr;

  PetscFunctionBegin;
  /* Create a histogram of nonlinear iterates over time */
  ctx->nonlinearViewer = PETSC_NULL;
  ctx->nonlinearHist   = PETSC_NULL;
  ierr = PetscOptionsGetString(PETSC_NULL, "-nonlinear_viewer", viewerType, 5, &opt);                     CHKERRQ(ierr);
  if (opt) {
    bins = 4;
    ierr = PetscOptionsGetInt(PETSC_NULL, "-nonlinear_hist_bins", &bins, &opt);                           CHKERRQ(ierr);
    ierr = PetscStrcasecmp(viewerType, "null", &opt);                                                     CHKERRQ(ierr);
    if (opt == PETSC_TRUE) {
      ierr = PetscDrawOpenNull(ctx->comm, &draw);                                                         CHKERRQ(ierr);
      ierr = PetscDrawHGCreate(draw, bins, &ctx->nonlinearHist);                                          CHKERRQ(ierr);
      ierr = PetscDrawDestroy(draw);                                                                      CHKERRQ(ierr);
    } else {
      ierr = PetscViewerDrawOpen(ctx->comm, 0, "Nonlinear Iterations", 0, 0, 300, 300, &ctx->nonlinearViewer); CHKERRQ(ierr);
      ierr = PetscViewerDrawGetDraw(ctx->nonlinearViewer, 0, &draw);                                      CHKERRQ(ierr);
      ierr = PetscDrawHGCreate(draw, bins, &ctx->nonlinearHist);                                          CHKERRQ(ierr);
    }
    ierr = PetscDrawHGCalcStats(ctx->nonlinearHist, PETSC_TRUE);                                          CHKERRQ(ierr);
    ierr = PetscDrawHGIntegerBins(ctx->nonlinearHist, PETSC_TRUE);                                        CHKERRQ(ierr);
  }

  /* Create a histogram of secondary nonlinear iterates over time */
  ctx->secondaryViewer = PETSC_NULL;
  ctx->secondaryHist   = PETSC_NULL;
  if (ctx->useSecondary == PETSC_TRUE) {
    ierr = PetscOptionsGetString(PETSC_NULL, "-nonlinear_viewer", viewerType, 5, &opt);                   CHKERRQ(ierr);
    if (opt) {
      bins = 4;
      ierr = PetscOptionsGetInt(PETSC_NULL, "-nonlinear_hist_bins", &bins, &opt);                         CHKERRQ(ierr);
      ierr = PetscStrcasecmp(viewerType, "null", &opt);                                                   CHKERRQ(ierr);
      if (opt == PETSC_TRUE) {
        ierr = PetscDrawOpenNull(ctx->comm, &draw);                                                       CHKERRQ(ierr);
        ierr = PetscDrawHGCreate(draw, bins, &ctx->secondaryHist);                                        CHKERRQ(ierr);
        ierr = PetscDrawDestroy(draw);                                                                    CHKERRQ(ierr);
      } else {
        ierr = PetscViewerDrawOpen(ctx->comm, 0, "Nonlinear Iterations", 0, 0, 300, 300, &ctx->secondaryViewer); CHKERRQ(ierr);
        ierr = PetscViewerDrawGetDraw(ctx->secondaryViewer, 0, &draw);                                    CHKERRQ(ierr);
        ierr = PetscDrawHGCreate(draw, bins, &ctx->secondaryHist);                                        CHKERRQ(ierr);
      }
      ierr = PetscDrawHGCalcStats(ctx->secondaryHist, PETSC_TRUE);                                        CHKERRQ(ierr);
      ierr = PetscDrawHGIntegerBins(ctx->secondaryHist, PETSC_TRUE);                                      CHKERRQ(ierr);
    }
  }

  /* Create a histogram of linear iterates over time */
  ctx->linearViewer      = PETSC_NULL;
  ctx->linearHist        = PETSC_NULL;
  ctx->totalLinearViewer = PETSC_NULL;
  ctx->totalLinearHist   = PETSC_NULL;
  ierr = PetscOptionsGetString(PETSC_NULL, "-linear_viewer", viewerType, 5, &opt);                        CHKERRQ(ierr);
  if (opt) {
    bins = 4;
    ierr = PetscOptionsGetInt(PETSC_NULL, "-linear_hist_bins", &bins, &opt);                              CHKERRQ(ierr);
    ierr = PetscStrcasecmp(viewerType, "null", &opt);                                                     CHKERRQ(ierr);
    if (opt == PETSC_TRUE) {
      ierr = PetscDrawOpenNull(ctx->comm, &draw);                                                         CHKERRQ(ierr);
      ierr = PetscDrawHGCreate(draw, bins, &ctx->linearHist);                                             CHKERRQ(ierr);
      ierr = PetscDrawDestroy(draw);                                                                      CHKERRQ(ierr);
    } else {
      ierr = PetscViewerDrawOpen(ctx->comm, 0, "Linear Iterations", 350, 0, 300, 300, &ctx->linearViewer);CHKERRQ(ierr);
      ierr = PetscViewerDrawGetDraw(ctx->linearViewer, 0, &draw);                                         CHKERRQ(ierr);
      ierr = PetscDrawHGCreate(draw, bins, &ctx->linearHist);                                             CHKERRQ(ierr);
    }
    ierr = PetscDrawHGCalcStats(ctx->linearHist, PETSC_TRUE);                                             CHKERRQ(ierr);
    ierr = PetscDrawHGIntegerBins(ctx->linearHist, PETSC_TRUE);                                           CHKERRQ(ierr);
    bins = 4;
    ierr = PetscOptionsGetInt(PETSC_NULL, "-total_linear_hist_bins", &bins, &opt);                        CHKERRQ(ierr);
    ierr = PetscStrcasecmp(viewerType, "null", &opt);                                                     CHKERRQ(ierr);
    if (opt == PETSC_TRUE) {
      ierr = PetscDrawOpenNull(ctx->comm, &draw);                                                         CHKERRQ(ierr);
      ierr = PetscDrawHGCreate(draw, bins, &ctx->totalLinearHist);                                        CHKERRQ(ierr);
      ierr = PetscDrawDestroy(draw);                                                                      CHKERRQ(ierr);
    } else {
      ierr = PetscViewerDrawOpen(ctx->comm, 0, "Total Linear Iterations", 700, 0, 300, 300, &ctx->totalLinearViewer); CHKERRQ(ierr);
      ierr = PetscViewerDrawGetDraw(ctx->totalLinearViewer, 0, &draw);                                    CHKERRQ(ierr);
      ierr = PetscDrawHGCreate(draw, bins, &ctx->totalLinearHist);                                        CHKERRQ(ierr);
    }
    ierr = PetscDrawHGCalcStats(ctx->totalLinearHist, PETSC_TRUE);                                        CHKERRQ(ierr);
    ierr = PetscDrawHGIntegerBins(ctx->totalLinearHist, PETSC_TRUE);                                      CHKERRQ(ierr);
  }

  /* Create a histogram of inner linear iterates over time */
  ctx->innerLinearViewer = PETSC_NULL;
  ctx->innerLinearHist   = PETSC_NULL;
  ctx->schurLinearViewer = PETSC_NULL;
  ctx->schurLinearHist   = PETSC_NULL;
  ierr = PetscOptionsGetString(PETSC_NULL, "-inner_linear_viewer", viewerType, 5, &opt);                  CHKERRQ(ierr);
  if (opt && (ctx->useSchur == PETSC_TRUE)) {
    bins = 4;
    ierr = PetscOptionsGetInt(PETSC_NULL, "-inner_linear_hist_bins", &bins, &opt);                        CHKERRQ(ierr);
    ierr = PetscStrcasecmp(viewerType, "null", &opt);                                                     CHKERRQ(ierr);
    if (opt == PETSC_TRUE) {
      ierr = PetscDrawOpenNull(ctx->comm, &draw);                                                         CHKERRQ(ierr);
      ierr = PetscDrawHGCreate(draw, bins, &ctx->innerLinearHist);                                        CHKERRQ(ierr);
      ierr = PetscDrawDestroy(draw);                                                                      CHKERRQ(ierr);
    } else {
      ierr = PetscViewerDrawOpen(ctx->comm, 0, "Inner Linear Iterations", 700, 0, 300, 300, &ctx->innerLinearViewer); CHKERRQ(ierr);
      ierr = PetscViewerDrawGetDraw(ctx->innerLinearViewer, 0, &draw);                                    CHKERRQ(ierr);
      ierr = PetscDrawHGCreate(draw, bins, &ctx->innerLinearHist);                                        CHKERRQ(ierr);
    }
    ierr = PetscDrawHGCalcStats(ctx->innerLinearHist, PETSC_TRUE);                                        CHKERRQ(ierr);
    ierr = PetscDrawHGIntegerBins(ctx->innerLinearHist, PETSC_TRUE);                                      CHKERRQ(ierr);
    bins = 4;
    ierr = PetscOptionsGetInt(PETSC_NULL, "-schur_linear_hist_bins", &bins, &opt);                        CHKERRQ(ierr);
    ierr = PetscStrcasecmp(viewerType, "null", &opt);                                                     CHKERRQ(ierr);
    if (opt == PETSC_TRUE) {
      ierr = PetscDrawOpenNull(ctx->comm, &draw);                                                         CHKERRQ(ierr);
      ierr = PetscDrawHGCreate(draw, bins, &ctx->schurLinearHist);                                        CHKERRQ(ierr);
      ierr = PetscDrawDestroy(draw);                                                                      CHKERRQ(ierr);
    } else {
      ierr = PetscViewerDrawOpen(ctx->comm, 0, "Schur Linear Iterations", 700, 0, 300, 300, &ctx->schurLinearViewer); CHKERRQ(ierr);
      ierr = PetscViewerDrawGetDraw(ctx->schurLinearViewer, 0, &draw);                                    CHKERRQ(ierr);
      ierr = PetscDrawHGCreate(draw, bins, &ctx->schurLinearHist);                                        CHKERRQ(ierr);
    }
    ierr = PetscDrawHGCalcStats(ctx->schurLinearHist, PETSC_TRUE);                                        CHKERRQ(ierr);
    ierr = PetscDrawHGIntegerBins(ctx->schurLinearHist, PETSC_TRUE);                                      CHKERRQ(ierr);
  }
  PetscFunctionReturn(0);
}

#undef  __FUNCT__
#define __FUNCT__ "NavierStokesResetHists"
/*@
  NavierStokesResetHists - This function clears problem histograms, perhaps printing them first.

  Collective on NavierStokesContext

  Input Parameters:
+ ctx   - The NavierStokesContext
- print - The flag to print the histograms

  Level: advanced

.seealso NavierStokesDestroyHists(), NavierStokesUpdateLinearHists()
@*/
int NavierStokesResetHists(NavierStokesContext ctx, PetscTruth print)
{
  int ierr;

  PetscFunctionBegin;
  /* Output nonlinear iteration histogram */
  if (ctx->nonlinearHist != PETSC_NULL) {
    if (print == PETSC_TRUE) {
      PetscPrintf(ctx->comm, "Histogram of Nonlinear Iteration Counts\n");
      ierr = PetscDrawHGPrint(ctx->nonlinearHist);                                                        CHKERRQ(ierr);
    }
    ierr = PetscDrawHGReset(ctx->nonlinearHist);                                                          CHKERRQ(ierr);
  }
  /* Output secondary nonlinear iteration histogram */
  if (ctx->secondaryHist != PETSC_NULL) {
    if (print == PETSC_TRUE) {
      PetscPrintf(ctx->comm, "Histogram of Secondary Nonlinear Iteration Counts\n");
      ierr = PetscDrawHGPrint(ctx->secondaryHist);                                                        CHKERRQ(ierr);
    }
    ierr = PetscDrawHGReset(ctx->secondaryHist);                                                          CHKERRQ(ierr);
  }
  /* Output linear iteration histograms */
  if (ctx->linearHist != PETSC_NULL) {
    if (print == PETSC_TRUE) {
      PetscPrintf(ctx->comm, "Histogram of Linear Iteration Counts\n");
      ierr = PetscDrawHGPrint(ctx->linearHist);                                                           CHKERRQ(ierr);
      PetscPrintf(ctx->comm, "Histogram of Total Linear Iteration Counts\n");
      ierr = PetscDrawHGPrint(ctx->totalLinearHist);                                                      CHKERRQ(ierr);
    }
    ierr = PetscDrawHGReset(ctx->linearHist);                                                             CHKERRQ(ierr);
    ierr = PetscDrawHGReset(ctx->totalLinearHist);                                                        CHKERRQ(ierr);
  }
  /* Output inner linear iteration histograms */
  if (ctx->innerLinearHist != PETSC_NULL) {
    if (print == PETSC_TRUE) {
      PetscPrintf(ctx->comm, "Histogram of Inner Linear Iteration Counts\n");
      ierr = PetscDrawHGPrint(ctx->innerLinearHist);                                                      CHKERRQ(ierr);
      PetscPrintf(ctx->comm, "Histogram of Schur Linear Iteration Counts\n");
      ierr = PetscDrawHGPrint(ctx->schurLinearHist);                                                      CHKERRQ(ierr);
    }
    ierr = PetscDrawHGReset(ctx->innerLinearHist);                                                        CHKERRQ(ierr);
    ierr = PetscDrawHGReset(ctx->schurLinearHist);                                                        CHKERRQ(ierr);
  }
  PetscFunctionReturn(0);
}

#undef  __FUNCT__
#define __FUNCT__ "NavierStokesDestroyHists"
/*@
  NavierStokesDestroyHists - This function destroys problem histograms.

  Collective on NavierStokesContext

  Input Parameter:
. ctx   - The NavierStokesContext

  Level: advanced

.seealso NavierStokesResetHists(), NavierStokesUpdateLinearHists()
@*/
int NavierStokesDestroyHists(NavierStokesContext ctx)
{
  int ierr;

  PetscFunctionBegin;
  /* Destroy nonlinear iteration histogram */
  if (ctx->nonlinearHist != PETSC_NULL) {
    if (ctx->nonlinearViewer != PETSC_NULL) {
      ierr = PetscViewerDestroy(ctx->nonlinearViewer);                                                    CHKERRQ(ierr);
    }
    ierr = PetscDrawHGDestroy(ctx->nonlinearHist);                                                        CHKERRQ(ierr);
  }
  /* Destroy secondary nonlinear iteration histogram */
  if (ctx->secondaryHist != PETSC_NULL) {
    if (ctx->secondaryViewer != PETSC_NULL) {
      ierr = PetscViewerDestroy(ctx->secondaryViewer);                                                    CHKERRQ(ierr);
    }
    ierr = PetscDrawHGDestroy(ctx->secondaryHist);                                                        CHKERRQ(ierr);
  }
  /* Destroy linear iteration histograms */
  if (ctx->linearHist != PETSC_NULL) {
    if (ctx->linearViewer != PETSC_NULL) {
      ierr = PetscViewerDestroy(ctx->linearViewer);                                                       CHKERRQ(ierr);
    }
    ierr = PetscDrawHGDestroy(ctx->linearHist);                                                           CHKERRQ(ierr);
    if (ctx->totalLinearViewer != PETSC_NULL) {
      ierr = PetscViewerDestroy(ctx->totalLinearViewer);                                                  CHKERRQ(ierr);
    }
    ierr = PetscDrawHGDestroy(ctx->totalLinearHist);                                                      CHKERRQ(ierr);
  }
  /* Destroy inner linear iteration histograms */
  if (ctx->innerLinearHist != PETSC_NULL) {
    if (ctx->innerLinearViewer != PETSC_NULL) {
      ierr = PetscViewerDestroy(ctx->innerLinearViewer);                                                  CHKERRQ(ierr);
    }
    ierr = PetscDrawHGDestroy(ctx->innerLinearHist);                                                      CHKERRQ(ierr);
    if (ctx->schurLinearViewer != PETSC_NULL) {
      ierr = PetscViewerDestroy(ctx->schurLinearViewer);                                                  CHKERRQ(ierr);
    }
    ierr = PetscDrawHGDestroy(ctx->schurLinearHist);                                                      CHKERRQ(ierr);
  }
  PetscFunctionReturn(0);
}

#undef  __FUNCT__
#define __FUNCT__ "NavierStokesUpdateLinearHists"
/*@
  NavierStokesUpdateLinearHists - This function updates the linear convergence histograms.

  Collective on GSNES

  Input Parameters:
+ sles - The solver context
- ctx  - The NavierStokesContext

  Level: advanced

.seealso NavierStokesUpdateGSNES()
@*/
int NavierStokesUpdateLinearHists(SLES sles, NavierStokesContext ctx)
{
  KSP ksp;
  PC  pc;
  int its, schurIts;
  int ierr;

  PetscFunctionBegin;
  ierr = SLESGetKSP(sles, &ksp);                                                                          CHKERRQ(ierr);
  ierr = SLESGetPC(sles, &pc);                                                                            CHKERRQ(ierr);
  /* Update linear iterations histogram */
  ierr = KSPGetIterationNumber(ksp, &its);                                                                CHKERRQ(ierr);
  if (ctx->linearHist != PETSC_NULL) {
    if (its > 0) {
      ierr = PetscDrawHGAddValue(ctx->linearHist, its);                                                   CHKERRQ(ierr);
    }
    ierr = PetscDrawHGDraw(ctx->linearHist);                                                              CHKERRQ(ierr);
  }
  /* Update inner linear iterations histogram */
  if (ctx->useSchur == PETSC_TRUE) {
    ierr = PCSchurGetIterationNumber(pc, &its, &schurIts);                                                CHKERRQ(ierr);
  }
  if (ctx->innerLinearHist != PETSC_NULL) {
    if (its > 0) {
      ierr = PetscDrawHGAddValue(ctx->innerLinearHist, its);                                              CHKERRQ(ierr);
    }
    ierr = PetscDrawHGDraw(ctx->innerLinearHist);                                                         CHKERRQ(ierr);
  }
  if (ctx->schurLinearHist != PETSC_NULL) {
    if (schurIts > 0) {
      ierr = PetscDrawHGAddValue(ctx->schurLinearHist, schurIts);                                         CHKERRQ(ierr);
    }
    ierr = PetscDrawHGDraw(ctx->schurLinearHist);                                                         CHKERRQ(ierr);
  }
  PetscFunctionReturn(0);
}

#undef  __FUNCT__
#define __FUNCT__ "NavierStokesUpdateGSNES"
/*@
  NavierStokesUpdateGSNES - This function updates the preconditioners and convergence histograms.

  Collective on GSNES

  Input Parameters:
+ snes - The GSNES context
- step - The current iteration

.seealso NavierStokesUpdateGTS()
@*/
int NavierStokesUpdateGSNES(GSNES snes, int step)
{
  NavierStokesContext ctx;
  SLES                sles;
  int                 ierr;

  PetscFunctionBegin;
  ierr = SNESGetJacobian(snes, PETSC_NULL, PETSC_NULL, (void **) &ctx, PETSC_NULL);                       CHKERRQ(ierr);
  ierr = SNESGetSLES(snes, &sles);                                                                        CHKERRQ(ierr);
  ierr = NavierStokesUpdateLinearHists(sles, ctx);                                                        CHKERRQ(ierr);
  /* Update multilevel information */
  if (ctx->useML == PETSC_TRUE) {
    ierr = PCMultiLevelUpdateSNES(snes, step);                                                            CHKERRQ(ierr);
  }
  PetscFunctionReturn(0);
}

#undef  __FUNCT__
#define __FUNCT__ "NavierStokesAnalyzeSolvers"
/*@
  NavierStokesAnalyzeSolvers - This function performs any online analysis or checking
  of the iterative solvers used for the linear and nonlinear problems.

  Collective on NavierStokesContext

  Input Parameter:
. ctx  - The NavierStokes context

.seealso NavierStokesUpdateGSNES()
@*/
int NavierStokesAnalyzeSolvers(NavierStokesContext ctx) {
  SLES sles;
  int  its, linearIts, secondaryIts, numFailures;
  int  ierr;

  PetscFunctionBegin;
  if (ctx->useConvection == PETSC_TRUE) {
    ierr = SNESGetIterationNumber(ctx->snes, &its);                                                       CHKERRQ(ierr);
    ierr = SNESGetNumberLinearIterations(ctx->snes, &linearIts);                                          CHKERRQ(ierr);
    ierr = SNESGetNumberUnsuccessfulSteps(ctx->snes, &numFailures);                                       CHKERRQ(ierr);
    PetscPrintf(ctx->comm, "%d failures occurred in the nonlinear iteration\n", numFailures);
    if (ctx->useSecondary) {
      ierr = SNESGetIterationNumber(ctx->snesSecondary, &secondaryIts);                                   CHKERRQ(ierr);
      ierr = SNESGetNumberLinearIterations(ctx->snesSecondary, &linearIts);                               CHKERRQ(ierr);
      ierr = SNESGetNumberUnsuccessfulSteps(ctx->snesSecondary, &numFailures);                            CHKERRQ(ierr);
      PetscPrintf(ctx->comm, "%d failures occurred in the secondary nonlinear iteration\n", numFailures);
    }

    /* Update nonlinear iterations histogram */
    if (ctx->nonlinearHist != PETSC_NULL) {
      if (its > 0) {
        ierr = PetscDrawHGAddValue(ctx->nonlinearHist, its);                                              CHKERRQ(ierr);
      }
      ierr = PetscDrawHGDraw(ctx->nonlinearHist);                                                         CHKERRQ(ierr);
    }
    /* Update secondary nonlinear iterations histogram */
    if (ctx->secondaryHist != PETSC_NULL) {
      if (secondaryIts > 0) {
        ierr = PetscDrawHGAddValue(ctx->secondaryHist, secondaryIts);                                     CHKERRQ(ierr);
      }
      ierr = PetscDrawHGDraw(ctx->secondaryHist);                                                         CHKERRQ(ierr);
    }
    /* Update total linear iterations histogram */
    if (ctx->totalLinearHist != PETSC_NULL) {
      if (linearIts > 0) {
        ierr = PetscDrawHGAddValue(ctx->totalLinearHist, linearIts);                                      CHKERRQ(ierr);
      }
      ierr = PetscDrawHGDraw(ctx->totalLinearHist);                                                       CHKERRQ(ierr);
    }
    /* Update linear iterations histogram */
    ierr = SNESGetSLES(ctx->snes, &sles);                                                                 CHKERRQ(ierr);
    ierr = NavierStokesUpdateLinearHists(sles, ctx);                                                      CHKERRQ(ierr);
  } else {
    ierr = NavierStokesUpdateLinearHists(ctx->sles, ctx);                                                 CHKERRQ(ierr);
  }
  PetscFunctionReturn(0);
}

/*-------------------------------------------------- Operators -------------------------------------------------------*/
#undef  __FUNCT__
#define __FUNCT__ "ConvectionOperator"
/* This defines the nonlinear operator

     u \cdot \nabla u

   and when used as an ALE operator, it gives

     (u - u_m) \cdot \nabla u
*/
int ConvectionOperator(int n, int comp, double *x, double *y, double *z, int numArgs, PetscScalar **fieldVals, PetscScalar *values, void *ctx)
{
  int i, c, d;

  PetscFunctionBegin;
  if (numArgs != 2) SETERRQ(PETSC_ERR_ARG_OUTOFRANGE, "Two arguments are needed for u \\cdot \\nabla u");
  for (i = 0; i < n; i++) {
    for(c = 0; c < comp; c++) {
      values[i*comp+c] = 0.0;
      for(d = 0; d < comp; d++)
#if 0
        values[i*comp+c] += fieldVals[0][d*3]*fieldVals[1][c*3+(d+1)];
#else
        values[i*comp+c] += 0.5*(fieldVals[0][d*3]*fieldVals[1][c*3+(d+1)] + fieldVals[1][d*3]*fieldVals[0][c*3+(d+1)]);
#endif
    }
  }
  PetscLogFlops(n*comp*comp*2);
  PetscFunctionReturn(0);
}

#undef  __FUNCT__
#define __FUNCT__ "LinearConvectionOperator"
/*
  This is a linearization of the convection operator $u \cdot \nabla u$:

    LC(u)^{cc'}_{ij} = \int_{E} \phi^c_i \phi^{c'}_j \PartDer{u^c}{x^{c'}}
                              + \delta_{cc'} \phi^c_i \sum_{c''} u^{c''} \PartDer{\phi^c_j}{x^{c''}}

  where $i,j$ are node indices and $c,c'$ are component indices, which together run over all variables.
*/
int LinearConvectionOperator(Discretization disc, Discretization test, int rowSize, int colSize,
                             int globalRowStart, int globalColStart, int globalSize, double *coords,
                             PetscScalar alpha, PetscScalar *field, PetscScalar *array, void *ctx)
{
  PetscScalar *fieldVal;          /* Field value and derivatives at a quadrature point */
  int     numQuadPoints;     /* Number of points used for Gaussian quadrature */
  double *quadPoints;        /* Points in the standard element for Gaussian quadrature */
  double *quadWeights;       /* Weights in the standard element for Gaussian quadrature */
  double *quadShapeFuncs;    /* Shape functions evaluated at quadrature points */
  double *quadTestFuncs;     /* Test  functions evaluated at quadrature points */
  double *quadShapeFuncDers; /* Shape function derivatives evaluated at quadrature points */
  double  dxxi;              /* \PartDer{x}{\xi}  */
  double  dxet;              /* \PartDer{x}{\eta} */
  double  dyxi;              /* \PartDer{y}{\xi}  */
  double  dyet;              /* \PartDer{y}{\eta} */
  double  dxix;              /* \PartDer{\xi}{x}  */
  double  detx;              /* \PartDer{\eta}{x} */
  double  dxiy;              /* \PartDer{\xi}{y}  */
  double  dety;              /* \PartDer{\eta}{y} */
  double  dphix;             /* \PartDer{\phi_i}{x} */
  double  dphiy;             /* \PartDer{\phi_i}{y} */
  double  dfxi;              /* \PartDer{field}{\xi}  */
  double  dfet;              /* \PartDer{field}{\eta} */
  double  jac;               /* |J| for map to standard element */
  double  invjac;            /* |J^{-1}| for map from standard element */
  int     comp;              /* The number of field components */
  int     funcs;             /* The number of shape functions */
  PetscScalar  nu;                /* The kinematic viscosity */
  int     i, j, c, d, f, p;
  int     ierr;

  PetscFunctionBegin;
  /* Calculate element matrix entries by Gaussian quadrature */
  nu   = *((PetscScalar *) ctx);
  ierr = DiscretizationGetNumComponents(disc, &comp);                                                     CHKERRQ(ierr);
  ierr = DiscretizationGetNumFunctions(disc, &funcs);                                                     CHKERRQ(ierr);
  ierr = DiscretizationGetFieldValues(disc, &fieldVal);                                                   CHKERRQ(ierr);
  ierr = DiscretizationGetNumQuadraturePoints(disc, &numQuadPoints);                                      CHKERRQ(ierr);
  ierr = DiscretizationGetQuadraturePoints(disc, &quadPoints);                                            CHKERRQ(ierr);
  ierr = DiscretizationGetQuadratureWeights(disc, &quadWeights);                                          CHKERRQ(ierr);
  ierr = DiscretizationGetQuadratureFunctions(disc, &quadShapeFuncs);                                     CHKERRQ(ierr);
  ierr = DiscretizationGetQuadratureDerivatives(disc, &quadShapeFuncDers);                                CHKERRQ(ierr);
  ierr = DiscretizationGetQuadratureFunctions(test, &quadTestFuncs);                                      CHKERRQ(ierr);
  for(p = 0; p < numQuadPoints; p++) {
    /* \PartDer{x}{\xi}(p)  = \sum^{funcs}_{f=1} x_f \PartDer{\phi^f(p)}{\xi}
       \PartDer{x}{\eta}(p) = \sum^{funcs}_{f=1} x_f \PartDer{\phi^f(p)}{\eta}
       \PartDer{y}{\xi}(p)  = \sum^{funcs}_{f=1} y_f \PartDer{\phi^f(p)}{\xi}
       \PartDer{y}{\eta}(p) = \sum^{funcs}_{f=1} y_f \PartDer{\phi^f(p)}{\eta} */
    dxxi = 0.0; dxet = 0.0;
    dyxi = 0.0; dyet = 0.0;
    for(j = 0; j < comp*3; j++) fieldVal[j] = 0.0;
    for(f = 0; f < funcs; f++) {
      dxxi += coords[f*2]  *quadShapeFuncDers[p*funcs*2+f*2];
      dxet += coords[f*2]  *quadShapeFuncDers[p*funcs*2+f*2+1];
      dyxi += coords[f*2+1]*quadShapeFuncDers[p*funcs*2+f*2];
      dyet += coords[f*2+1]*quadShapeFuncDers[p*funcs*2+f*2+1];
      for(j = 0; j < comp; j++) {
        fieldVal[j*3]   += field[f*comp+j]*quadShapeFuncs[p*funcs+f];
        fieldVal[j*3+1] += field[f*comp+j]*quadShapeFuncDers[p*funcs*2+f*2];
        fieldVal[j*3+2] += field[f*comp+j]*quadShapeFuncDers[p*funcs*2+f*2+1];
      }
    }
    jac = fabs(dxxi*dyet - dxet*dyxi);
#ifdef PETSC_USE_BOPT_g
    if (jac < 1.0e-14) {
      PetscPrintf(PETSC_COMM_SELF, "p: %d x1: %lf y1: %lf x2: %lf y2: %lf x3: %lf y3: %lf\n",
                  p, coords[0], coords[1], coords[2], coords[3], coords[4], coords[5]);
      SETERRQ(PETSC_ERR_DISC_SING_JAC, "Singular Jacobian");
    }
#endif
    /* These are the elements of the inverse matrix */
    invjac = 1.0/jac;
    dxix   =  dyet*invjac;
    dxiy   = -dxet*invjac;
    detx   = -dyxi*invjac;
    dety   =  dxxi*invjac;

    /* Convert the field derivatives to old coordinates */
    for(j = 0; j < comp; j++) {
      dfxi            = fieldVal[j*3+1];
      dfet            = fieldVal[j*3+2];
      fieldVal[j*3+1] = dfxi*dxix + dfet*detx;
      fieldVal[j*3+2] = dfxi*dxiy + dfet*dety;
    }

    for(i = 0; i < funcs; i++) {
      for(j = 0; j < funcs; j++) {
        for(c = 0; c < comp; c++) {
          for(d = 0; d < comp; d++) {
            array[(i*comp+c+globalRowStart)*globalSize + j*comp+d+globalColStart] +=
              alpha*quadTestFuncs[p*funcs+i]*quadShapeFuncs[p*funcs+j]*fieldVal[c*3+(d+1)]*jac*quadWeights[p];
            if (c == d) {
              dphix = quadShapeFuncDers[p*funcs*2+j*2]*dxix + quadShapeFuncDers[p*funcs*2+j*2+1]*detx;
              dphiy = quadShapeFuncDers[p*funcs*2+j*2]*dxiy + quadShapeFuncDers[p*funcs*2+j*2+1]*dety;
              array[(i*comp+c+globalRowStart)*globalSize + j*comp+d+globalColStart] +=
                alpha*quadTestFuncs[p*funcs+i]*(fieldVal[0*3]*dphix + fieldVal[1*3]*dphiy)*jac*quadWeights[p];
            }
          }
        }
      }
    }
  }
  PetscLogFlops((8*funcs + 7*funcs*comp + 8 + 6*funcs*funcs*comp*comp + 14*funcs*funcs*comp) * numQuadPoints);

  PetscFunctionReturn(0);
}

/* This defines the linear operator

     {1 \over 2} \nabla \cdot ( \nabla u + {\nabla u}^T )
*/
#undef  __FUNCT__
#define __FUNCT__ "RateOfStrainTensor"
int RateOfStrainTensor(Discretization disc, Discretization test, int rowSize, int colSize,
                       int globalRowStart, int globalColStart, int globalSize, double *coords,
                       PetscScalar alpha, PetscScalar *field, PetscScalar *array, void *ctx)
{
  int     comp;              /* The number of field components */
  int     funcs;             /* The number of shape functions */
  int     numQuadPoints;     /* Number of points used for Gaussian quadrature */
  double *quadWeights;       /* Weights in the standard element for Gaussian quadrature */
  double *quadShapeFuncDers; /* Shape function derivatives evaluated at quadrature points */
  double *quadTestFuncDers;  /* Test  function derivatives evaluated at quadrature points */
  double  dxxi;              /* \PartDer{x}{\xi}  */
  double  dxet;              /* \PartDer{x}{\eta} */
  double  dyxi;              /* \PartDer{y}{\xi}  */
  double  dyet;              /* \PartDer{y}{\eta} */
  double  dxix;              /* \PartDer{\xi}{x}  */
  double  detx;              /* \PartDer{\eta}{x} */
  double  dxiy;              /* \PartDer{\xi}{y}  */
  double  dety;              /* \PartDer{\eta}{y} */
  double  dphix;             /* \PartDer{\phi_j}{x} Shape */
  double  dphiy;             /* \PartDer{\phi_j}{y} Shape */
  double  dpsix;             /* \PartDer{\psi_i}{x} Test  */
  double  dpsiy;             /* \PartDer{\psi_i}{y} Test  */
  double  jac;               /* |J| for map to standard element */
  double  invjac;            /* |J^{-1}| for map from standard element */
  int     i, j, f, p;
  int     ierr;

  PetscFunctionBegin;
  ierr = DiscretizationGetNumComponents(disc, &comp);                                                     CHKERRQ(ierr);
  ierr = DiscretizationGetNumFunctions(disc, &funcs);                                                     CHKERRQ(ierr);
  if (comp != 2) SETERRQ(PETSC_ERR_ARG_WRONG, "Operator only valid with 2D vector field");

  ierr = DiscretizationGetNumQuadraturePoints(disc, &numQuadPoints);                                      CHKERRQ(ierr);
  ierr = DiscretizationGetQuadratureWeights(disc, &quadWeights);                                          CHKERRQ(ierr);
  ierr = DiscretizationGetQuadratureDerivatives(disc, &quadShapeFuncDers);                                CHKERRQ(ierr);
  ierr = DiscretizationGetQuadratureDerivatives(test, &quadTestFuncDers);                                 CHKERRQ(ierr);

  /* Bring out a factor of 1/2 */
  alpha *= 0.5;

  /* Calculate element matrix entries by Gaussian quadrature */
  for(p = 0; p < numQuadPoints; p++) {
    /* \PartDer{x}{\xi}(p)  = \sum^{funcs}_{f=1} x_f \PartDer{\phi^f(p)}{\xi}
       \PartDer{x}{\eta}(p) = \sum^{funcs}_{f=1} x_f \PartDer{\phi^f(p)}{\eta}
       \PartDer{y}{\xi}(p)  = \sum^{funcs}_{f=1} y_f \PartDer{\phi^f(p)}{\xi}
       \PartDer{y}{\eta}(p) = \sum^{funcs}_{f=1} y_f \PartDer{\phi^f(p)}{\eta} */
    dxxi = 0.0; dxet = 0.0;
    dyxi = 0.0; dyet = 0.0;
    for(f = 0; f < funcs; f++) {
      dxxi += coords[f*2]  *quadShapeFuncDers[p*funcs*2+f*2];
      dxet += coords[f*2]  *quadShapeFuncDers[p*funcs*2+f*2+1];
      dyxi += coords[f*2+1]*quadShapeFuncDers[p*funcs*2+f*2];
      dyet += coords[f*2+1]*quadShapeFuncDers[p*funcs*2+f*2+1];
    }
    jac  = fabs(dxxi*dyet - dxet*dyxi);
#ifdef PETSC_USE_BOPT_g
    if (jac < 1.0e-14) {
      PetscPrintf(PETSC_COMM_SELF, "p: %d x1: %lf y1: %lf x2: %lf y2: %lf x3: %lf y3: %lf\n",
                  p, coords[0], coords[1], coords[2], coords[3], coords[4], coords[5]);
      SETERRQ(PETSC_ERR_DISC_SING_JAC, "Singular Jacobian");
    }
#endif
    /* These are the elements of the inverse matrix */
    invjac =  1.0/jac;
    dxix   =  dyet*invjac;
    dxiy   = -dxet*invjac;
    detx   = -dyxi*invjac;
    dety   =  dxxi*invjac;

    for(i = 0; i < funcs; i++) {
      for(j = 0; j < funcs; j++) {
        dpsix = quadTestFuncDers[p*funcs*2+i*2] *dxix + quadTestFuncDers[p*funcs*2+i*2+1] *detx;
        dphix = quadShapeFuncDers[p*funcs*2+j*2]*dxix + quadShapeFuncDers[p*funcs*2+j*2+1]*detx;
        dpsiy = quadTestFuncDers[p*funcs*2+i*2] *dxiy + quadTestFuncDers[p*funcs*2+i*2+1] *dety;
        dphiy = quadShapeFuncDers[p*funcs*2+j*2]*dxiy + quadShapeFuncDers[p*funcs*2+j*2+1]*dety;
        array[(i*comp+0+globalRowStart)*globalSize + j*comp+0+globalColStart] +=
          -alpha*(2.0*dpsix*dphix +     dpsiy*dphiy)*jac*quadWeights[p];
        array[(i*comp+0+globalRowStart)*globalSize + j*comp+1+globalColStart] +=
          -alpha*(    dpsiy*dphix                  )*jac*quadWeights[p];
        array[(i*comp+1+globalRowStart)*globalSize + j*comp+0+globalColStart] +=
          -alpha*(                      dpsix*dphiy)*jac*quadWeights[p];
        array[(i*comp+1+globalRowStart)*globalSize + j*comp+1+globalColStart] +=
          -alpha*(    dpsix*dphix + 2.0*dpsiy*dphiy)*jac*quadWeights[p];
      }
    }
  }
  PetscLogFlops((8*funcs + 8 + 38*funcs*funcs) * numQuadPoints);

  PetscFunctionReturn(0);
}

/*---------------------------------------------- Main Computation ---------------------------------------------------*/
#undef  __FUNCT__
#define __FUNCT__ "NavierStokesSolve"
int NavierStokesSolve(NavierStokesContext ctx, GVec f, GVec u, int *its) {
  PC  pc;
  int ierr;

  PetscFunctionBegin;
  /* Primary solve */
  if (ctx->useConvection == PETSC_TRUE) {
    ierr = SNESSolve(ctx->snes, u, its);                                                                  CHKERRQ(ierr);
  } else {
    /* Solve P^T_2 A P_2 (P^{-1} u)_2 = P^T_2 (f - A P_1 D^{-1} Z^T g) */
    ierr = SLESSolve(ctx->sles, f, u, its);                                                               CHKERRQ(ierr);

    if (ctx->useML == PETSC_TRUE) {
      /* Calculate p = Z D^{-1} P^T_1 (f - A P_2 (P^{-1} u)_2 - A P_1 D^{-1} Z^T g) */
      ierr = SLESGetPC(ctx->sles, &pc);                                                                   CHKERRQ(ierr);
      ierr = PCMultiLevelGetMultiplier(pc, u, ctx->p);                                                    CHKERRQ(ierr);

      /* Recover u */
      ierr = PCMultiLevelBuildSolution(pc, u);                                                            CHKERRQ(ierr);
    }

    /* Show pressure */
    ierr = GVecViewFromOptions(ctx->pExact, "Exact Pressure");                                            CHKERRQ(ierr);
    ierr = GVecViewFromOptions(ctx->p,      "Pressure");                                                  CHKERRQ(ierr);
  }
  /* Show solution */
  ierr = GVecViewFromOptions(ctx->uExact, "Exact Solution");                                              CHKERRQ(ierr);
  ierr = GVecViewFromOptions(ctx->u,      "Solution");                                                    CHKERRQ(ierr);
  if (ctx->mathViewer != PETSC_NULL) {
    ierr = GVecView(u, ctx->mathViewer);                                                                  CHKERRQ(ierr);
  }

  /* Secondary solve */
  if ((ctx->useConvection == PETSC_TRUE) && (ctx->useSecondary == PETSC_TRUE)) {
    PetscPrintf(ctx->comm, "Solving with secondary solver\n");
    ierr = SNESSolve(ctx->snesSecondary, u, its);                                                         CHKERRQ(ierr);
    /* Show solution */
    ierr = GVecViewFromOptions(ctx->u, "SecondarySolution");                                              CHKERRQ(ierr);
    if (ctx->mathViewer != PETSC_NULL) {
      ierr = GVecView(u, ctx->mathViewer);                                                                CHKERRQ(ierr);
    }
  }
  PetscFunctionReturn(0);
}

#undef  __FUNCT__
#define __FUNCT__ "NavierStokesComputeFlowField"
int NavierStokesComputeFlowField(NavierStokesContext ctx) {
  int its;  /* The iteration count for the linear solver */
  int loop;
  int ierr;

  PetscLogStagePush(UserStages[Stage_Setup]);
  ierr = NavierStokesSetup(ctx);                                                                          CHKERRQ(ierr);
  PetscLogStagePop();

  for(loop = 0; loop < ctx->numLoops; loop++) {
    if (loop >= ctx->refineStep) {
      PetscLogStagePush(UserStages[Stage_GridRefine]);
      ierr = NavierStokesRefineGrid(ctx);                                                                 CHKERRQ(ierr);
      PetscLogStagePop();
    }

    PetscLogStagePush(UserStages[NumStages+loop]);
    /* Setup problem */
    ierr = NavierStokesSetupGrid(ctx);                                                                    CHKERRQ(ierr);
    ierr = NavierStokesCreateStructures(ctx);                                                             CHKERRQ(ierr);
    ierr = NavierStokesSetupStructures(ctx);                                                              CHKERRQ(ierr);

    /* Check the exact solution */
    ierr = NavierStokesCheckSolution(ctx, ctx->uExact, ctx->pExact, "exact");                             CHKERRQ(ierr);

    /* Solve system */
    ierr = NavierStokesSolve(ctx, ctx->f, ctx->u, &its);                                                  CHKERRQ(ierr);

    /* Check the computed solution */
    ierr = NavierStokesCheckSolution(ctx, ctx->u, ctx->p, "computed");                                    CHKERRQ(ierr);

    /* Log results */
    ierr = NavierStokesAnalyzeSolvers(ctx);                                                               CHKERRQ(ierr);

    /* Cleanup */
    ierr = NavierStokesDestroyStructures(ctx);                                                            CHKERRQ(ierr);
    PetscLogStagePop();
  }
  PetscFunctionReturn(0);
}
